# -*- coding: utf-8 -*-
# Copyright 2021-2022 releng-tool

from releng_tool.defs import GlobalAction
from releng_tool.util.io import execute
from releng_tool.util.io import generate_temp_dir
from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import touch
from tests.support.site_tool_test import TestSiteToolBase
import os

DEFAULT_BRANCH = 'test'

class TestToolGit(TestSiteToolBase):
    def prepare_defconfig(self, defconfig):
        self.defconfig_add('VCS_TYPE', 'git')

    def prepare_repo_dir(self, repo_dir):
        self._git_repo('init', repo_dir, '--initial-branch=' + DEFAULT_BRANCH)
        self._git_repo('config', 'user.email', 'test@releng.io', repo=repo_dir)
        self._git_repo('config', 'user.name', 'Unit Test', repo=repo_dir)
        self._create_commit('initial commit', repo=repo_dir)

    def test_tool_git_basic_branch(self):
        self.defconfig_add('VERSION', DEFAULT_BRANCH)
        rv = self.engine.run()
        self.assertTrue(rv)

    def test_tool_git_basic_commit(self):
        hash_ = self._create_commit()
        self.defconfig_add('VERSION', hash_)
        rv = self.engine.run()
        self.assertTrue(rv)

    def test_tool_git_basic_tag(self):
        tag = self._create_tag('my-tag')
        self.defconfig_add('VERSION', tag)
        rv = self.engine.run()
        self.assertTrue(rv)

    def test_tool_git_branch_forward_slash(self):
        # create a commit on a new branch with a forward slash
        TMP_BRANCH = 'container/test'
        self._git_repo('checkout', '-b', TMP_BRANCH)
        self._create_commit()
        self._git_repo('checkout', DEFAULT_BRANCH)

        self.defconfig_add('VERSION', TMP_BRANCH)
        rv = self.engine.run()
        self.assertTrue(rv)

    def test_tool_git_changing_revision(self):
        # verify pulling from an original tag
        first_hash = self._create_commit('first commit')
        first_tag = self._create_tag('first')

        self.defconfig_add('VERSION', first_tag)
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, first_hash)

        # cleanup
        self.cleanup_outdir()

        # create a new remote revision, update the target and see if it uses the
        # newer tag
        second_hash = self._create_commit('second commit')
        second_tag = self._create_tag('second')

        self.defconfig_add('VERSION', second_tag)
        rv = self.engine.run()
        self.assertTrue(rv)

        # cleanup
        self.cleanup_outdir()

        # update the package definition to use the older tag and see if we can
        # still load the older tag properly
        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, second_hash)

        self.defconfig_add('VERSION', first_tag)
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, first_hash)

    def test_tool_git_custom_options_invalid(self):
        self.defconfig_add('GIT_CONFIG', {
            'bad-config-key': '',
        })
        self.defconfig_add('VERSION', DEFAULT_BRANCH)
        rv = self.engine.run()
        self.assertFalse(rv)

    def test_tool_git_custom_options_valid(self):
        MY_CUSTOM_CFG = 'tmp.example'
        MY_CUSTOM_CFG_VALUE = 'my-value'

        self.defconfig_add('GIT_CONFIG', {
            MY_CUSTOM_CFG: MY_CUSTOM_CFG_VALUE,
        })
        self.defconfig_add('VERSION', DEFAULT_BRANCH)
        rv = self.engine.run()
        self.assertTrue(rv)

        fetched_cfg = self._git_cache('config', '--get', MY_CUSTOM_CFG)
        self.assertEqual(fetched_cfg, MY_CUSTOM_CFG_VALUE)

    def test_tool_git_custom_reference(self):
        # create a commit on a new custom reference
        TMP_BRANCH = 'tmp'
        self._git_repo('checkout', '-b', TMP_BRANCH)
        hash_ = self._create_commit()
        self._git_repo('update-ref', 'refs/mycustomref/test', hash_)
        self._git_repo('checkout', DEFAULT_BRANCH)
        self._git_repo('branch', '-D', TMP_BRANCH)

        # first fetch attempt should fail since the custom reference will not be
        # fetched by default
        self.defconfig_add('VERSION', hash_)
        rv = self.engine.run()
        self.assertFalse(rv)

        # add the new custom reference to the configuration and re-fetch
        self.defconfig_add('GIT_REFSPECS', ['mycustomref/*'])
        rv = self.engine.run()
        self.assertTrue(rv)

    def test_tool_git_depth_shallow_default(self):
        self._create_commit()
        hash_ = self._create_commit()
        self.defconfig_add('VERSION', hash_)

        rv = self.engine.run()
        self.assertTrue(rv)

        rev_count = self._git_cache('rev-list', '--count', 'HEAD')
        self.assertEqual(int(rev_count), 1)

    def test_tool_git_depth_shallow_disabled_config(self):
        self._create_commit()
        hash_ = self._create_commit()
        self.defconfig_add('GIT_DEPTH', 0)
        self.defconfig_add('VERSION', hash_)

        rv = self.engine.run()
        self.assertTrue(rv)

        rev_count = self._git_cache('rev-list', '--count', 'HEAD')
        self.assertNotEqual(int(rev_count), 1)

    def test_tool_git_depth_shallow_disabled_quirk(self):
        self.engine.opts.quirks.append('releng.git.no_depth')

        self._create_commit()
        hash_ = self._create_commit()
        self.defconfig_add('VERSION', hash_)

        rv = self.engine.run()
        self.assertTrue(rv)

        rev_count = self._git_cache('rev-list', '--count', 'HEAD')
        self.assertNotEqual(int(rev_count), 1)

    def test_tool_git_depth_unshallow_initial(self):
        hash_ = self._create_commit()
        self._create_commit()
        self.defconfig_add('VERSION', hash_)

        rv = self.engine.run()
        self.assertTrue(rv)

        rev_count = self._git_cache('rev-list', '--count', 'HEAD')
        self.assertNotEqual(int(rev_count), 1)

    def test_tool_git_depth_unshallow_updated(self):
        self._create_commit()
        commit_a = self._create_commit()
        commit_b = self._create_commit()
        self.defconfig_add('VERSION', commit_b)

        # the engine run should fetch a single commit
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, commit_b)
        rev_count = self._git_cache('rev-list', '--count', 'HEAD')
        self.assertEqual(int(rev_count), 1)

        # cleanup
        self.cleanup_outdir()

        # update to a older commit, check if the engine fetch will unshallow the
        # currently shallow state cache
        self.defconfig_add('VERSION', commit_a)

        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, commit_a)
        rev_count = self._git_cache('rev-list', '--count', 'HEAD')
        self.assertNotEqual(int(rev_count), 1)

    def test_tool_git_devmode_auto_fetch_branch(self):
        # create a new branch with a commit
        TARGET_BRANCH = 'target'
        self._git_repo('checkout', '-b', TARGET_BRANCH)
        first_hash = self._create_commit('first')

        # enable development mode
        touch(self.engine.opts.ff_devmode)

        # note: default state should be automatic cache-ignoring
        #        (i.e enabled in development mode)

        # fetch the target
        self.defconfig_add('VERSION', TARGET_BRANCH)
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, first_hash)

        # cleanup
        self.cleanup_outdir()
        self.engine.opts.devmode = None

        # create a new commit
        second_hash = self._create_commit('second')

        # extract project; should be the second commit
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, second_hash)

        # cleanup
        self.cleanup_outdir()
        self.engine.opts.devmode = None

        # create a new commit
        third_hash = self._create_commit('third')

        # explicitly indicate that cache-ignoring should not happen
        self.defconfig_add('DEVMODE_IGNORE_CACHE', False)
        self.defconfig_dump()

        # extract; should be the second commit
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, second_hash)

        # cleanup
        self.cleanup_outdir()
        self.engine.opts.devmode = None

        # explicitly indicate that cache-ignoring can happen
        self.defconfig_add('DEVMODE_IGNORE_CACHE', True)

        # extract; should be the third commit
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, third_hash)

    def test_tool_git_diverged_branch(self):
        # create a new branch with a commit
        TARGET_BRANCH = 'target'
        self._git_repo('checkout', '-b', TARGET_BRANCH)
        first_hash = self._create_commit('first')

        # fetch the target
        self.defconfig_add('VERSION', TARGET_BRANCH)
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, first_hash)

        # cleanup
        self.cleanup_outdir()

        # destroy the old branch and build a new one with the same name
        self._git_repo('checkout', DEFAULT_BRANCH)
        self._git_repo('branch', '-D', TARGET_BRANCH)
        self._git_repo('checkout', '-b', TARGET_BRANCH)
        second_hash = self._create_commit('second')

        # force a fetch (which should update the cache)
        self.engine.opts.gbl_action = GlobalAction.FETCH
        rv = self.engine.run()
        self.assertTrue(rv)

        # fetch/extract the new target
        self.engine.opts.gbl_action = GlobalAction.EXTRACT
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, second_hash)

    def test_tool_git_remote_tag_change(self):
        tag_name = 'my-tag'

        first_hash = self._create_commit('first')
        self._create_tag(tag_name)
        self.defconfig_add('VERSION', tag_name)

        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, first_hash)

        # cleanup
        self.cleanup_outdir()

        # re-invoke with a new tag on a remote should not change cache state,
        # as we should not have fetched
        self._delete_tag(tag_name)
        second_hash = self._create_commit('second')
        self._create_tag(tag_name)

        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, first_hash)

        # cleanup
        self.cleanup_outdir()

        # force a fetch (which should update the cache)
        self.engine.opts.gbl_action = GlobalAction.FETCH
        rv = self.engine.run()
        self.assertTrue(rv)

        # re-invoke extraction stage to see if it uses the newer tagged commit
        self.engine.opts.gbl_action = GlobalAction.EXTRACT
        rv = self.engine.run()
        self.assertTrue(rv)

        current_hash = self._git_cache('rev-parse', 'HEAD')
        self.assertEqual(current_hash, second_hash)

    def test_tool_git_submodules_default(self):
        self.defconfig_add('VERSION', DEFAULT_BRANCH)

        with generate_temp_dir() as repo2:
            # prepare additional mock repository directories
            repo1 = self.repo_dir
            self.prepare_repo_dir(repo2)

            # dummy file on repo1
            touch(os.path.join(repo1, 'file1'))
            self._create_commit('add file', repo=repo1, add=True)

            # dummy file on repo2
            touch(os.path.join(repo2, 'file2'))
            self._create_commit('add file', repo=repo2, add=True)

            # add a submodule repo2 to repo1
            self._git_repo('submodule', 'add', repo2, 'repo2', repo=repo1)
            self._create_commit('add module', repo=repo1, add=True)

            # extract package but not submodules (by default)
            self.engine.opts.gbl_action = GlobalAction.EXTRACT
            rv = self.engine.run()
            self.assertTrue(rv)

            # verify expected content from main package; not submodules
            out_dir = os.path.join(
                self.engine.opts.build_dir, 'test-' + DEFAULT_BRANCH)
            repo1_file = os.path.join(out_dir, 'file1')
            repo2_file = os.path.join(out_dir, 'repo2', 'file2')
            self.assertTrue(os.path.exists(repo1_file))
            self.assertFalse(os.path.exists(repo2_file))

    def test_tool_git_submodules_enabled(self):
        self.defconfig_add('GIT_SUBMODULES', True)
        self.defconfig_add('VERSION', DEFAULT_BRANCH)

        with generate_temp_dir() as repo2, generate_temp_dir() as repo3:
            # prepare additional mock repository directories
            repo1 = self.repo_dir
            self.prepare_repo_dir(repo2)
            self.prepare_repo_dir(repo3)

            # dummy file on repo1
            touch(os.path.join(repo1, 'file1'))
            self._create_commit('add file', repo=repo1, add=True)

            # dummy file on repo2
            touch(os.path.join(repo2, 'file2'))
            self._create_commit('add file', repo=repo2, add=True)

            # dummy file on repo3
            touch(os.path.join(repo3, 'file3'))
            self._create_commit('add file', repo=repo3, add=True)

            # add a submodule repo2 to repo1
            self._git_repo('submodule', 'add', repo2, 'repo2', repo=repo1)
            self._create_commit('add module', repo=repo1, add=True)

            # add a submodule repo3 to repo2
            self._git_repo('submodule', 'add', repo3, 'repo3', repo=repo2)
            self._create_commit('add module', repo=repo2, add=True)

            # extract package and submodules
            self.engine.opts.gbl_action = GlobalAction.EXTRACT
            rv = self.engine.run()
            self.assertTrue(rv)

            # verify expected content from main package and submodules
            out_dir = os.path.join(
                self.engine.opts.build_dir, 'test-' + DEFAULT_BRANCH)
            repo1_file = os.path.join(out_dir, 'file1')
            repo2_file = os.path.join(out_dir, 'repo2', 'file2')
            repo3_file = os.path.join(out_dir, 'repo2', 'repo3', 'file3')
            self.assertTrue(os.path.exists(repo1_file))
            self.assertTrue(os.path.exists(repo2_file))
            self.assertTrue(os.path.exists(repo3_file))

    def test_tool_git_submodules_branch_revision(self):
        self.defconfig_add('GIT_SUBMODULES', True)
        self.defconfig_add('VERSION', DEFAULT_BRANCH)

        with generate_temp_dir() as repo2:
            # prepare additional mock repository directories
            repo1 = self.repo_dir
            self.prepare_repo_dir(repo2)

            # dummy file on repo1
            touch(os.path.join(repo1, 'file1'))
            self._create_commit('add file', repo=repo1, add=True)

            # dummy files on repo2 for two branches
            CUSTOM_BRANCH = 'custom'

            touch(os.path.join(repo2, 'file2'))
            self._create_commit('add file', repo=repo2, add=True)

            self._git_repo('checkout', '-b', CUSTOM_BRANCH, repo=repo2)

            touch(os.path.join(repo2, 'file3'))
            self._create_commit('add file', repo=repo2, add=True)

            self._git_repo('checkout', DEFAULT_BRANCH, repo=repo2)

            # add a submodule repo2 to repo1
            self._git_repo('submodule', 'add', repo2, 'repo2', repo=repo1)
            self._create_commit('add module', repo=repo1, add=True)

            # extract package but not submodules (by default)
            self.engine.opts.gbl_action = GlobalAction.EXTRACT
            rv = self.engine.run()
            self.assertTrue(rv)

            # verify expected content from main package; not submodules
            out_dir = os.path.join(
                self.engine.opts.build_dir, 'test-' + DEFAULT_BRANCH)
            repo1_file = os.path.join(out_dir, 'file1')
            repo2_file = os.path.join(out_dir, 'repo2', 'file2')
            self.assertTrue(os.path.exists(repo1_file))
            self.assertTrue(os.path.exists(repo2_file))

            # cleanup
            self.cleanup_outdir()

            # adjust submodule to target the custom branch
            self._git_repo('config', '-f', '.gitmodules',
                'submodule.repo2.branch', CUSTOM_BRANCH, repo=repo1)

            # force a fetch (which should update the cache)
            self.engine.opts.gbl_action = GlobalAction.FETCH
            rv = self.engine.run()
            self.assertTrue(rv)

            # re-extract package and submodules
            self.engine.opts.gbl_action = GlobalAction.EXTRACT
            rv = self.engine.run()
            self.assertTrue(rv)

            # verify expected content from main package and submodules
            repo3_file = os.path.join(out_dir, 'repo2', 'file3')
            self.assertTrue(os.path.exists(repo3_file))

    def test_tool_git_unknown_branch_tag(self):
        self.defconfig_add('VERSION', 'unknown')
        rv = self.engine.run()
        self.assertFalse(rv)

    def test_tool_git_unknown_hash(self):
        self.defconfig_add('VERSION',
            'da39a3ee5e6b4b0d3255bfef95601890afd80709')
        rv = self.engine.run()
        self.assertFalse(rv)

    def _git(self, workdir, *args):
        with interim_working_dir(workdir):
            out = []
            if not execute(['git'] + list(args), capture=out, critical=False):
                print(['git'] + list(args))
                print('\n'.join(out))
                assert False, 'failed to issue git command'
            return '\n'.join(out)

    def _git_cache(self, *args):
        return self._git(self.cache_dir, *args)

    def _git_repo(self, *args, **kwargs):
        repo = kwargs.get('repo')
        if not repo:
            repo = self.repo_dir
        return self._git(repo, *args)

    def _create_commit(self, msg='test', **kwargs):
        repo = kwargs.get('repo')
        if kwargs.get('add'):
            self._git_repo('add', '.', repo=repo)
        self._git_repo('commit', '--allow-empty', '-m', msg, repo=repo)
        return self._git_repo('rev-parse', 'HEAD', repo=repo)

    def _create_tag(self, tag, repo=None):
        self._git_repo('tag', tag, repo=repo)
        return tag

    def _delete_tag(self, tag, repo=None):
        self._git_repo('tag', '-d', tag, repo=repo)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.hg import HG
from releng_tool.util.log import err
from releng_tool.util.log import log

def extract(opts):
    """
    support extraction (checkout) of a mercurial cache into a build directory

    With provided extraction options (``RelengExtractOptions``), the extraction
    stage will be processed. A Mercurial extraction process will populate a
    working tree based off the cached Mercurial repository acquired from the
    fetch stage.

    Args:
        opts: the extraction options

    Returns:
        ``True`` if the extraction stage is completed; ``False`` otherwise
    """

    assert opts
    cache_dir = opts.cache_dir
    revision = opts.revision
    work_dir = opts.work_dir

    if not HG.exists():
        err('unable to extract package; mercurial (hg) is not installed')
        return None

    log('checking out target revision into work tree')
    if not HG.execute(['--verbose', 'clone', '--rev', revision,
            cache_dir, work_dir],
            cwd=work_dir):
        err('unable to checkout revision')
        return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from tests import prepare_testenv
import os
import unittest

class TestCliOverrides(unittest.TestCase):

    def test_cli_overrides(self):
        cfg = {
            'injected_kv': {
                'TEST_OVERRIDE': '123',
            },
        }
        expected_file_flag = 'override-detected'

        with prepare_testenv(config=cfg, template='override-check') as engine:
            engine.run()

            file_flag = os.path.join(engine.opts.target_dir, expected_file_flag)
            self.assertTrue(os.path.exists(file_flag))



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.util.enum import Enum
import hashlib

class PkgKeyType(Enum):
    """
    package key type

    Enumeration of types supported when fetching configuration values defined by
    a package definition.

    Attributes:
        BOOL: boolean value
        DICT: dictionary value
        DICT_STR_STR: dictionary of string pairs value
        DICT_STR_STR_OR_STRS: dictionary of string pairs or strings value
        STR: single string value
        STRS: one or more strings value
        INT_NONNEGATIVE: non-negative integer value
        INT_POSITIVE: positive integer value
    """
    BOOL = 'bool'
    DICT = 'dict'
    DICT_STR_STR = 'dict_str_str'
    DICT_STR_STR_OR_STRS = 'dict_str_str_or_strs'
    STR = 'str'
    STRS = 'strs'
    INT_NONNEGATIVE = 'int_nonnegative'
    INT_POSITIVE = 'int_positive'

def pkg_cache_key(site):
    """
    generate a cache key for a provided package's site

    Package's may share caching data if their sites match. This call returns a
    calculated "cache key" for a provided cache site.

    Returns:
        the cache key
    """
    return hashlib.sha1(site.encode('utf_8')).hexdigest()

def pkg_key(pkg, type_):
    """
    generate a package key for a given type string

    Generates a compatible "package key" for a unsanitized package name ``pkg``
    of a specific key ``type``. The package string is "cleaned" to replaces
    select characters (such as dashes) with underscores and becomes uppercase.
    For example, consider the package name "my-awesome-module". For a package
    key "VERSION", the complete key for this package is
    "MY_AWESOME_MODULE_VERSION".

    Args:
        pkg: the package name
        type_: the package key type

    Returns:
        the completed package key
    """
    clean = pkg
    for c in [' ', '*', '-', '.', ':', '?', '|']:
        clean = clean.replace(c, '_')
    return '{}_{}'.format(clean.upper(), type_)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.api import RelengBuildOptions
from releng_tool.defs import PackageType
from releng_tool.engine.autotools.build import build as build_autotools
from releng_tool.engine.cmake.build import build as build_cmake
from releng_tool.engine.python.build import build as build_python
from releng_tool.engine.script.build import build as build_script
from releng_tool.util import nullish_coalescing as NC
from releng_tool.util.api import replicate_package_attribs
from releng_tool.util.io import interim_working_dir
from releng_tool.util.log import err
from releng_tool.util.log import note
import sys

def stage(engine, pkg, script_env):
    """
    handles the build stage for a package

    With a provided engine and package instance, the build stage will be
    processed.

    Args:
        engine: the engine
        pkg: the package being built
        script_env: script environment information

    Returns:
        ``True`` if the build stage is completed; ``False`` otherwise
    """

    note('building {}...', pkg.name)
    sys.stdout.flush()

    if pkg.build_subdir:
        build_dir = pkg.build_subdir
    else:
        build_dir = pkg.build_dir

    build_opts = RelengBuildOptions()
    replicate_package_attribs(build_opts, pkg)
    build_opts.build_defs = pkg.build_defs
    build_opts.build_dir = build_dir
    build_opts.build_env = pkg.build_env
    build_opts.build_opts = pkg.build_opts
    build_opts.build_output_dir = pkg.build_output_dir
    build_opts.def_dir = pkg.def_dir
    build_opts.env = script_env
    build_opts.ext = pkg.ext_modifiers
    build_opts.host_dir = engine.opts.host_dir
    build_opts.name = pkg.name
    build_opts.prefix = NC(pkg.prefix, engine.opts.sysroot_prefix)
    build_opts.staging_dir = engine.opts.staging_dir
    build_opts.symbols_dir = engine.opts.symbols_dir
    build_opts.target_dir = engine.opts.target_dir
    build_opts.version = pkg.version
    build_opts._quirks = engine.opts.quirks

    # if package has a job-override value, use it over any global option
    if pkg.fixed_jobs:
        build_opts.jobs = pkg.fixed_jobs
        build_opts.jobsconf = pkg.fixed_jobs
    else:
        build_opts.jobs = engine.opts.jobs
        build_opts.jobsconf = engine.opts.jobsconf

    builder = None
    if pkg.type in engine.registry.package_types:
        def _(opts):
            return engine.registry.package_types[pkg.type].build(pkg.type, opts)
        builder = _
    elif pkg.type == PackageType.AUTOTOOLS:
        builder = build_autotools
    elif pkg.type == PackageType.CMAKE:
        builder = build_cmake
    elif pkg.type == PackageType.PYTHON:
        builder = build_python
    elif pkg.type == PackageType.SCRIPT:
        builder = build_script

    if not builder:
        err('build type is not implemented: {}', pkg.type)
        return False

    with interim_working_dir(build_dir):
        built = builder(build_opts)
        if not built:
            return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool

#: executable used to run make commands
MAKE_COMMAND = 'make'

#: make host tool helper
MAKE = RelengTool(MAKE_COMMAND)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.util import nullish_coalescing
import unittest

class TestUtilCommon(unittest.TestCase):
    def test_util_nullish_coalescing(self):
        # using provided value
        value = nullish_coalescing('test', None)
        self.assertEqual(value, 'test')

        value = nullish_coalescing('test2', 'test3')
        self.assertEqual(value, 'test2')

        # using default
        value = nullish_coalescing(None, 'test4')
        self.assertEqual(value, 'test4')

        # sanity check against False evaluated types
        checks = [
            '',
            0,
            dict(),
            list(),
            set(),
        ]
        for check in checks:
            value = nullish_coalescing(check, 'err')
            self.assertEqual(value, check)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.util.log import releng_log_configuration
from releng_tool.util.win32 import enable_ansi as enable_ansi_win32
from tests import RelengToolTestSuite
import argparse
import fnmatch
import os
import sys
import unittest

#: default verbosity for unit tests
DEFAULT_VERBOSITY = 2

#: directory name for the standard unit tests
UNIT_TESTS_DIRNAME = 'unit-tests'

def main():
    """
    process main for unit tests

    This method will prepare the test suite, load listed test classes and
    perform a run.
    """
    loader = unittest.TestLoader()
    suite = RelengToolTestSuite()

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--test-dir', default=UNIT_TESTS_DIRNAME)
    parser.add_argument('--unbuffered', '-U', action='store_true')
    parser.add_argument('--verbose', '-V', action='count', default=0)

    args, args_extra = parser.parse_known_args()

    # configure logging for unit test output
    buffered = not args.unbuffered
    verbosity = 0
    if args.verbose:
        try:
            verbosity = int(args.verbose)
            # ignore first verbose level for unbuffered mode
            if buffered:
                verbosity -= 1
        except ValueError:
            pass

        buffered = False

    if args.debug:
        buffered = False
        if not verbosity:
            verbosity = 1

    releng_log_configuration(args.debug, False, verbosity)

    # disable short descriptions
    unittest.TestCase.shortDescription = lambda x: None

    # support character sequences (for color output on win32 cmd)
    if sys.platform == 'win32':
        enable_ansi_win32()

    # discover unit tests
    test_base = os.path.dirname(os.path.realpath(__file__))
    tests_dir = os.path.join(test_base, args.test_dir)
    unit_tests = loader.discover(tests_dir)

    # check if a unit test name was provided
    target_test_name_pattern = None
    for arg in args_extra:
        if not arg.startswith('-'):
            target_test_name_pattern = arg
            break

    # register tests
    target_unit_tests = None
    if target_test_name_pattern:
        target_unit_tests, module_load_failure = find_tests(
            unit_tests, target_test_name_pattern)
        if target_unit_tests:
            print('running specific tests:')
            for test in target_unit_tests:
                print('    {}'.format(test.id()))
            sys.stdout.flush()
        else:
            print('ERROR: unable to find test with pattern: '
                '{}'.format(target_test_name_pattern))
            if not module_load_failure:
                sys.exit(1)

    if target_unit_tests:
        suite.addTests(target_unit_tests)
    else:
        suite.addTests(unit_tests)

    # invoke test suite
    runner = unittest.TextTestRunner(buffer=buffered,
        verbosity=DEFAULT_VERBOSITY)
    return 0 if runner.run(suite).wasSuccessful() else 1

def find_tests(entity, pattern):
    """
    search for a unit tests with a matching wildcard pattern

    Looks for the first 'unittest.case.TestCase' instance where its identifier
    matches the provided wildcard pattern.

    Args:
        entity: the unit test entity to search for a pattern on
        pattern: the pattern

    Returns:
        a 2-tuple (list of tests, flag if a module error is detected)
    """
    found = []
    module_load_failure = False

    if isinstance(entity, unittest.case.TestCase):
        if fnmatch.fnmatch(entity.id(), '*{}*'.format(pattern)):
            found.append(entity)
        elif 'ModuleImportFailure' in entity.id():
            module_load_failure = True
    elif isinstance(entity, unittest.TestSuite):
        for subentity in entity:
            tests, fail = find_tests(subentity, pattern)
            if tests:
                found.extend(tests)
            if fail:
                module_load_failure = True

    return found, module_load_failure

if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.svn import SVN
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import note
import os
import sys
import tarfile

def fetch(opts):
    """
    support fetching from svn sources

    With provided fetch options (``RelengFetchOptions``), the fetch stage will
    be processed.

    Args:
        opts: fetch options

    Returns:
        ``True`` if the fetch stage is completed; ``False`` otherwise
    """

    assert opts
    cache_file = opts.cache_file
    name = opts.name
    revision = opts.revision
    site = opts.site
    work_dir = opts.work_dir

    if not SVN.exists():
        err('unable to fetch package; svn is not installed')
        return None

    note('fetching {}...'.format(name))
    sys.stdout.flush()

    log('checking out sources')
    if not SVN.execute(['checkout', '-r', revision, site, work_dir],
            cwd=work_dir):
        err('unable to checkout module')
        return None

    log('caching sources')
    def svn_exclude(file):
        if file.endswith('.svn'):
            return True
        return False

    # ensure cache file's directory exists
    cache_dir = os.path.abspath(os.path.join(cache_file, os.pardir))
    if not ensure_dir_exists(cache_dir):
        return None

    with tarfile.open(cache_file, 'w:gz') as tar:
        tar.add(work_dir, arcname=name, exclude=svn_exclude)

    return cache_file



# -*- coding: utf-8 -*-
# Copyright 2021-2022 releng-tool

from releng_tool.tool import RelengTool
from releng_tool.tool.tar import TAR
from releng_tool.tool.tar import TAR_COMMAND
from tests import prepare_testenv
from tests.support import fetch_unittest_assets_dir
from tests.support.http_daemon import httpd_context
import os
import sys
import unittest

class TestSiteUrl(unittest.TestCase):
    def test_site_url_fetch_archive_tar_bz2_valid(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.tar.bz2'.format(host, port)

            httpd_assets = fetch_unittest_assets_dir('sample-files')
            archive = os.path.join(httpd_assets, 'sample-files.tar.bz2')

            with open(archive, 'rb') as f:
                data = f.read()

            httpd.rsp.append((200, data))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertTrue(rv)

                outdir = os.environ['MINIMAL_BUILD_DIR']
                stripped_file = os.path.join(outdir, 'bz2-file-container.txt')

                self.assertTrue(os.path.exists(stripped_file))

    def test_site_url_fetch_archive_tar_invalid(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.tgz'.format(host, port)

            httpd.rsp.append((200, b'not-an-archive'))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertFalse(rv)

    def test_site_url_fetch_archive_tar_valid_external_extract(self):
        if not TAR.exists():
            raise unittest.SkipTest('environment does not have a tar command')

        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.tar'.format(host, port)

            httpd_assets = fetch_unittest_assets_dir('sample-files')
            archive = os.path.join(httpd_assets, 'sample-files.tar')

            with open(archive, 'rb') as f:
                data = f.read()

            httpd.rsp.append((200, data))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertTrue(rv)

                outdir = os.environ['MINIMAL_BUILD_DIR']
                stripped_file = os.path.join(outdir, 'tar-file-container.txt')

                self.assertTrue(os.path.exists(stripped_file))

    def test_site_url_fetch_archive_tar_valid_internal_extract(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.tar'.format(host, port)

            httpd_assets = fetch_unittest_assets_dir('sample-files')
            archive = os.path.join(httpd_assets, 'sample-files.tar')

            with open(archive, 'rb') as f:
                data = f.read()

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                otes = TAR.exists()
                try:
                    # temporarily force a non-external tar command (if needed)
                    RelengTool.detected[TAR_COMMAND] = False

                    httpd.rsp.append((200, data))
                    rv = engine.run()
                finally:
                    RelengTool.detected[TAR_COMMAND] = otes

                self.assertTrue(rv)

                outdir = os.environ['MINIMAL_BUILD_DIR']
                stripped_file = os.path.join(outdir, 'tar-file-container.txt')

                self.assertTrue(os.path.exists(stripped_file))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                # redo tar fetch but with a no strip option
                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))
                    f.write('MINIMAL_STRIP_COUNT=0\n')

                try:
                    RelengTool.detected[TAR_COMMAND] = False

                    httpd.rsp.append((200, data))
                    rv = engine.run()
                finally:
                    RelengTool.detected[TAR_COMMAND] = otes

                self.assertTrue(rv)

                outdir = os.environ['MINIMAL_BUILD_DIR']
                sample_root_file = os.path.join(outdir, 'tar-file-root')
                container_dir = os.path.join(outdir, 'container')
                sample_container_file = os.path.join(
                    container_dir, 'tar-file-container.txt')

                self.assertTrue(os.path.exists(sample_root_file))
                self.assertTrue(os.path.exists(sample_container_file))

    def test_site_url_fetch_archive_tgz_valid(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.tgz'.format(host, port)

            httpd_assets = fetch_unittest_assets_dir('sample-files')
            archive = os.path.join(httpd_assets, 'sample-files.tgz')

            with open(archive, 'rb') as f:
                data = f.read()

            httpd.rsp.append((200, data))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertTrue(rv)

                outdir = os.environ['MINIMAL_BUILD_DIR']
                stripped_file = os.path.join(outdir, 'tgz-file-container.txt')

                self.assertTrue(os.path.exists(stripped_file))

    def test_site_url_fetch_archive_txz_valid(self):
        if sys.version_info < (3, 0):
            raise unittest.SkipTest('tar may not support xz; ignoring')

        if not TAR.exists():
            raise unittest.SkipTest('environment does not have a tar command')

        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.tar.xz'.format(host, port)

            httpd_assets = fetch_unittest_assets_dir('sample-files')
            archive = os.path.join(httpd_assets, 'sample-files.tar.xz')

            with open(archive, 'rb') as f:
                data = f.read()

            httpd.rsp.append((200, data))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertTrue(rv)

                outdir = os.environ['MINIMAL_BUILD_DIR']
                stripped_file = os.path.join(outdir, 'txz-file-container.txt')

                self.assertTrue(os.path.exists(stripped_file))

    def test_site_url_fetch_archive_zip_invalid(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.zip'.format(host, port)

            httpd.rsp.append((200, b'not-an-archive'))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertFalse(rv)

    def test_site_url_fetch_archive_zip_valid(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.zip'.format(host, port)

            httpd_assets = fetch_unittest_assets_dir('sample-files')
            archive = os.path.join(httpd_assets, 'sample-files.zip')

            with open(archive, 'rb') as f:
                data = f.read()

            httpd.rsp.append((200, data))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertTrue(rv)

    def test_site_url_fetch_file_http(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/test.txt'.format(host, port)

            httpd.rsp.append((200, b'Sample text file.'))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertTrue(rv)

    def test_site_url_fetch_file_https(self):
        with httpd_context(secure=True) as httpd:
            host, port = httpd.server_address
            site = 'https://localhost:{}/test.txt'.format(port)

            httpd.rsp.append((200, b'Sample text file.'))

            with prepare_testenv(template='https-self-signed') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir, 'package', 'pkg', 'pkg')

                with open(pkg_script, 'a') as f:
                    f.write('PKG_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertTrue(rv)

    def test_site_url_fetch_missing(self):
        with httpd_context() as httpd:
            host, port = httpd.server_address
            site = 'http://{}:{}/missing.txt'.format(host, port)

            httpd.rsp.append((404, None))

            with prepare_testenv(template='minimal') as engine:
                root_dir = engine.opts.root_dir
                pkg_script = os.path.join(root_dir,
                    'package', 'minimal', 'minimal')

                with open(pkg_script, 'a') as f:
                    f.write('MINIMAL_SITE="{}"\n'.format(site))

                rv = engine.run()
                self.assertFalse(rv)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from __future__ import unicode_literals
from io import open
from releng_tool.defs import GlobalAction
from releng_tool.util.io import generate_temp_dir
from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import path_remove
from tests import prepare_testenv
import os
import unittest

PKG_NAME = 'test'
PKG_DEFINITION = os.path.join('package', PKG_NAME, PKG_NAME)
PKG_CFG_PREFIX = 'TEST_'
DEFAULT_TEMPLATE = 'site-tool'

class TestSiteToolBase(unittest.TestCase):
    """
    site tool testing unit test base

    Provides a base class for unit testing site-specific tools supported by
    releng-tool. Using this base helps prepare an interim repository directory,
    as well as a releng-tool engine/environment with a package configuration
    configured to use the interim repository directory as a site location.

    A unit test can build/manipulate a repository's contents, modify a test
    package's configuration and invoke a prepared releng-tool engine to help
    sanity check that releng-tool handles various package options and repository
    states.
    """

    def run(self, result=None):
        """
        run the test

        Run the test, collecting the result into the TestResult object passed as
        result. See `unittest.TestCase.run()` for more details.

        Args:
            result (optional): the test result to populate
        """

        with generate_temp_dir() as repo_dir, interim_working_dir(repo_dir):
            self.repo_dir = repo_dir
            self.prepare_repo_dir(repo_dir)

            with prepare_testenv(template=self.tool_template()) as engine:
                opts = engine.opts

                opts.gbl_action = self.prepare_global_action()

                self.cache_dir = os.path.join(opts.cache_dir, PKG_NAME)
                self.defconfig = os.path.join(opts.root_dir, PKG_DEFINITION)
                self.engine = engine

                self.defconfig_add('SITE', self.repo_dir)
                self.prepare_defconfig(self.defconfig)

                super(TestSiteToolBase, self).run(result)

    def prepare_global_action(self):
        """
        hook invoked to prepare a default global action for tool testing

        By default, site tool testing will default to only performing up to
        the extraction stage; however, some unit testing may wish to use an
        alternative default. This hook provides a means to override the
        default action.

        Returns:
            the global action
        """

        # by default, only perform up to the extraction stage
        return GlobalAction.EXTRACT

    def prepare_defconfig(self, defconfig):
        """
        hook invoked to prepare a package's definition

        Method invoked to allow an implementer help prepare a package's
        definition file to be used by a releng-tool engine.

        Args:
            defconfig: the definition file
        """

    def prepare_repo_dir(self, repo_dir):
        """
        hook invoked to prepare a repository

        Method invoked to allow an implementer help prepare a mocked
        repository's directory contents.

        Note that this call is invoked where the working directory is configured
        to the same path provided by `repo_dir`.

        Args:
            repo_dir: the repository directory
        """

    def tool_template(self):
        """
        the configuration template to use

        Returns the configuration template to be used when preparing a test
        environment for a unit test. This call is to be overridden by
        implementations using TestSiteToolBase.

        Returns:
            the template name; `None` if no template is to be used
        """
        return DEFAULT_TEMPLATE

    def cleanup_outdir(self):
        """
        cleanup the output directory

        Provides a unit test the ability to quickly cleanup the output directory
        of the configured output directory of a test.
        """
        path_remove(self.engine.opts.out_dir)

    def defconfig_add(self, key, value):
        """
        inject a configuration into the test package's definition

        This method will inject a configuration entry named by `key` and append
        it to the running test's package definition. The representation of the
        provided value will be used as an assignment for the configuration key.

        Args:
            key: the key to use for the configuration entry
            value: the value to use for the configuration entry
        """

        with open(self.defconfig, mode='a', encoding='utf_8') as file_def:
            file_def.write('{}{} = {}\n'.format(
                PKG_CFG_PREFIX, key, repr(value)))

    def defconfig_dump(self):
        """
        dump the test package's definition contents to standard out

        When requested, the test package's definition file will be read and
        dumped to the configured standard output stream.
        """
        with open(self.defconfig, mode='r', encoding='utf_8') as f:
            content = f.readlines()
        print('-------------------------------')
        print(''.join(content).strip())
        print('-------------------------------')

    def dir_dump(self, dir_):
        """
        dump a test folders's file list to standard out

        When requested, a provide directory will be scanned for files and the
        file list will be dumped to the configured standard output stream.

        Args:
            dir_: the directory to search
        """

        files = self._dir_fetch(dir_)

        print('DIR: {}]'.format(dir_))
        print('-------------------------------')
        for file_ in files:
            print(file_[len(dir_) + 1:])
        print('-------------------------------')

    def _dir_fetch(self, dir_):
        """
        compile a list of files in a test folder

        This call will return a list of all files found in the provided
        directory. If the provided directory contains other directories, they
        will also be searched for file entries to populate.

        Args:
            dir_: the directory to search

        Returns:
            the file list
        """

        files = []

        for f in os.scandir(dir_):
            if f.is_dir():
                files.extend(self._dir_fetch(f.path))
            elif f.is_file():
                files.append(f.path)

        return files



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from collections import OrderedDict
from datetime import datetime
from releng_tool import __version__ as releng_version
from releng_tool.defs import ConfKey
from releng_tool.defs import GlobalAction
from releng_tool.defs import PkgAction
from releng_tool.defs import VcsType
from releng_tool.engine.fetch import stage as fetch_stage
from releng_tool.engine.init import initialize_sample
from releng_tool.exceptions import RelengToolMissingConfigurationError
from releng_tool.exceptions import RelengToolMissingPackagesError
from releng_tool.opts import RELENG_POST_BUILD_NAME
from releng_tool.packages.exceptions import RelengToolStageFailure
from releng_tool.packages.manager import RelengPackageManager
from releng_tool.packages.pipeline import RelengPackagePipeline
from releng_tool.prerequisites import RelengPrerequisites
from releng_tool.registry import RelengRegistry
from releng_tool.stats import RelengStats
from releng_tool.support import releng_include
from releng_tool.support import require_version
from releng_tool.util.env import env_value
from releng_tool.util.env import extend_script_env
from releng_tool.util.file_flags import FileFlag
from releng_tool.util.file_flags import check_file_flag
from releng_tool.util.file_flags import process_file_flag
from releng_tool.util.io import FailedToPrepareWorkingDirectoryError
from releng_tool.util.io import cat
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.io import execute
from releng_tool.util.io import execute_rv
from releng_tool.util.io import generate_temp_dir
from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import ls
from releng_tool.util.io import opt_file
from releng_tool.util.io import path_copy
from releng_tool.util.io import path_exists
from releng_tool.util.io import path_move
from releng_tool.util.io import path_remove
from releng_tool.util.io import run_script
from releng_tool.util.io import touch
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import note
from releng_tool.util.log import success
from releng_tool.util.log import verbose
from releng_tool.util.log import warn
from releng_tool.util.platform import platform_exit
from releng_tool.util.string import expand
from releng_tool.util.string import interpret_dictionary_strings
from releng_tool.util.string import interpret_string
from releng_tool.util.string import interpret_strings
from shutil import copyfileobj
import os
import ssl
import sys

class RelengEngine:
    """
    releng tool's engine

    The "engine" invokes the main components of a releng process such as
    downloading, configuring, building and more.

    Args:
        opts: options used to configure the engine

    Attributes:
        opts: options used to configure the engine
        pkgman: manager for package-related tasks
        registry: extension registry
    """
    def __init__(self, opts):
        self.registry = RelengRegistry()
        self.opts = opts
        self.pkgman = RelengPackageManager(opts, self.registry, dvcs_cache=True)
        self.stats = RelengStats(opts)

    def run(self):
        """
        run the engine

        Starts the actual processing of the release engineering steps. When this
        method is returned, the process is completed.

        Returns:
            ``True`` if the engine has completed without error; ``False`` if an
            issue has occurred when interpreting or running the user's
            configuration/package definitions
        """

        opts = self.opts
        gaction = opts.gbl_action
        license_files = {}
        pa = opts.pkg_action

        if gaction == GlobalAction.INIT:
            return initialize_sample(opts)

        start_time = datetime.now().replace(microsecond=0)
        gbls = {
            'releng_args': opts.forward_args,
            'releng_version': releng_version,
        }

        # if any injected key-value entries are provided, attempt to add them
        # into the global context
        if opts.injected_kv:
            debug('injecting key-value arguments into global context')
            gbls.update(opts.injected_kv)

        debug('loading statistics...')
        self.stats.load()

        # inform the user of any active running modes
        if self.opts.devmode:
            verbose('running in development mode')

        if self.opts.local_srcs:
            verbose('running in local-sources mode')

        # register the project's root directory as a system path; permits a
        # project to import locally created modules in their build/etc. scripts
        debug('registering root directory in path...')
        sys.path.append(opts.root_dir)

        # register the project's host directory as a system path; lazily permits
        # loading host tools (not following prefix or bin container) built by a
        # project over the system
        debug('registering host directory in path...')
        sys.path.insert(0, opts.host_dir)
        os.environ['PATH'] = opts.host_dir + os.pathsep + os.environ['PATH']

        # prepare script environment to make helpers available to configuration
        # script(s)
        #
        # Note that some options prepared for the configuration script may
        # need to be updated after running the project's configuration (e.g.
        # possibly needing to update the `PREFIX` variable if `sysroot_prefix`
        # was configured). This is peformed later in the engine run call.
        self._prepare_script_environment(gbls, gaction, opts.pkg_action)

        verbose("loading user's configuration...")
        conf_point, conf_point_exists = opt_file(opts.conf_point)
        if not conf_point_exists:
            raise RelengToolMissingConfigurationError(conf_point)

        settings = run_script(conf_point, gbls, subject='configuration')
        if not settings:
            return False

        script_env = gbls.copy()
        extend_script_env(script_env, settings)
        self.pkgman.script_env = script_env
        verbose('configuration file loaded')

        # configuration overrides file for builders
        if os.path.isfile(opts.conf_point_overrides):
            warn('detected configuration overrides file')

            overrides = run_script(opts.conf_point_overrides, gbls,
                subject='configuration overrides')
            if not overrides:
                return False

            extend_script_env(script_env, overrides)
            extend_script_env(settings, overrides)
            verbose('configuration overrides file loaded')

        # handle cleaning requests
        if (gaction == GlobalAction.CLEAN or gaction == GlobalAction.MRPROPER
                or gaction == GlobalAction.DISTCLEAN):
            self._handle_clean_request(gaction)
            return True

        # file flag processing
        state = self._process_file_flags()
        if state is not None:
            debug('file-flag processing has triggered closure')
            return state

        # determine package name(s) extraction
        #
        # Compile a (minimum) list of package names to be processed; either from
        # the user explicitly provided package target or from the user's
        # configuration file.
        if opts.target_action:
            pkg_names = [opts.target_action]
        else:
            pkg_names = self._get_package_names(settings)
        if not pkg_names:
            return False
        debug('target packages)')
        for pkg_name in pkg_names:
            debug(' {}', pkg_name)

        # processing additional settings
        if not self._process_settings(settings):
            return False

        # register the project's host-bin directory as a system path; lazily
        # permits loading host tools built by a project over the system
        debug('registering host bin directory in path...')
        host_bin_dir = os.path.join(opts.host_dir + opts.sysroot_prefix, 'bin')
        sys.path.insert(0, host_bin_dir)
        os.environ['PATH'] = host_bin_dir + os.pathsep + os.environ['PATH']

        # load and process packages
        pkgs = self.pkgman.load(pkg_names)
        if not pkgs:
            return False

        # if cleaning a package, remove it's build output directory and stop
        if pa in (PkgAction.CLEAN, PkgAction.DISTCLEAN):
            for pkg in pkgs:
                if pkg.name == opts.target_action:
                    def pkg_verbose_clean(desc):
                        verbose('{} for package: {}', desc, pkg.name)
                    pkg_verbose_clean('removing output directory')
                    path_remove(pkg.build_output_dir)

                    if pa == PkgAction.DISTCLEAN:
                        if os.path.exists(pkg.cache_file):
                            pkg_verbose_clean('removing cache file')
                            path_remove(pkg.cache_file)
                        if os.path.isdir(pkg.cache_dir):
                            pkg_verbose_clean('removing cache directory')
                            path_remove(pkg.cache_dir)

                    return True
            assert False # should not reach here

        # ensure any of required host tools do exist
        if 'releng.disable_prerequisites_check' not in opts.quirks:
            prerequisites = RelengPrerequisites(pkgs, opts.prerequisites)
            if not prerequisites.check():
                return False

        # track if this action is "pre-configuration", where a package
        # dependency chain can be ignored
        requested_preconfig = pa in [
            PkgAction.EXTRACT,
            PkgAction.FETCH,
            PkgAction.LICENSE,
            PkgAction.PATCH,
        ]

        # determine if an explicit fetch request has been made
        requested_fetch = None
        if gaction == GlobalAction.FETCH or pa == PkgAction.FETCH:
            requested_fetch = True

        try:
            # ensure all package sources are acquired first
            for pkg in pkgs:
                if not self._stage_init(pkg):
                    return False

                # if this is a package-specific pre-configure action, only
                # ensure a fetched this specific package
                if requested_preconfig and pkg.name != opts.target_action:
                    continue

                # none/local-vcs-type packages do not need to fetch
                if pkg.vcs_type in (VcsType.LOCAL, VcsType.NONE):
                    continue

                # in the event that we are not explicit fetching and the package
                # has already been extracted, completely skip the fetching stage
                is_local_pkg = pkg.is_internal and opts.local_srcs
                if not requested_fetch and not is_local_pkg:
                    flag = pkg._ff_extract
                    if check_file_flag(flag) == FileFlag.EXISTS:
                        continue

                self.stats.track_duration_start(pkg.name, 'fetch')
                fetched = fetch_stage(
                    self, pkg, requested_fetch, pkg.fetch_opts)
                self.stats.track_duration_end(pkg.name, 'fetch')
                if not fetched:
                    return False

            # re-apply script environment to ensure previous script environment
            # changes have not manipulated the environment (from standard
            # helpers).
            self._prepare_script_environment(
                script_env, gaction, opts.pkg_action)

            # process each package (configuring, building, etc.)
            if gaction != GlobalAction.FETCH and pa != PkgAction.FETCH:

                # pre-populate "common" directories
                #
                # The following will attempt to pre-populate common directories
                # before attempting to process a package pipeline. A package
                # configured in this pipeline may wish to populate any one of
                # these directories anytime between a configuration stage to a
                # post-package stage. Ensuring these directories exist ahead of
                # time can help prevent scenarios where developers try to file
                # a file into a directory, but instead create a file as the
                # common directory name (e.g. `TARGET_DIR` points to a file).
                common_dirs = [
                    opts.host_dir,
                    opts.staging_dir,
                    opts.symbols_dir,
                    opts.target_dir,
                ]
                for common_dir in common_dirs:
                    if not ensure_dir_exists(common_dir):
                        return False

                pipeline = RelengPackagePipeline(self, opts, script_env)
                for pkg in pkgs:
                    # if this is a package-specific pre-configure action, only
                    # process the specific action
                    if requested_preconfig and pkg.name != opts.target_action:
                        continue

                    verbose('processing package: {}', pkg.name)
                    if not pipeline.process(pkg):
                        return True
                license_files = pipeline.license_files

        except RelengToolStageFailure:
            return False
        except FailedToPrepareWorkingDirectoryError as e:
            err('''\
unable to prepare a package's working directory

An attempt to prepare and move into a working directory for a package process
has failed. Ensure the following path is accessible for this user:

    {}''', e)
            return False

        is_action = (gaction or pa or opts.target_action is not None)

        # perform license generation
        if gaction == GlobalAction.LICENSES or pa == PkgAction.LICENSE \
                or not is_action:
            note('generating license information...')

            if not self._perform_license_generation(license_files):
                return False

        # perform post-processing and completion message if not performing a
        # specific action
        if not is_action:
            if not self._post_processing(script_env):
                err('failed to perform post-processing')
                return False

            end_time = datetime.now().replace(microsecond=0)
            success('completed ({})', end_time - start_time)

        # attempt to generate a report at the end of a run
        self.stats.generate()

        return True

    def _stage_init(self, pkg):
        """
        initialize the package environment for processing

        Performs initializations steps when processing a package. For example,
        tracking/pre-cleanup of package-specific file flags.

        Args:
            pkg: the package being processed

        Returns:
            ``True`` if the stage has been properly initialized; ``False``
            otherwise
        """

        # forced request
        #
        # When a forced request is made, clear all file flags so these stages
        # can be invoked again.
        if self.opts.force:
            path_remove(pkg._ff_bootstrap)
            path_remove(pkg._ff_configure)
            path_remove(pkg._ff_build)
            path_remove(pkg._ff_install)
            path_remove(pkg._ff_post)

        # user invoking a package-specific override
        #
        # If the user is invoking a package rebuild, reconfiguration, etc.,
        # ensure existing file flags are cleared to ensure these stages are
        # invoked again.
        elif pkg.name == self.opts.target_action:
            if (self.opts.pkg_action in
                    (PkgAction.REBUILD, PkgAction.REBUILD_ONLY)):
                path_remove(pkg._ff_build)
                path_remove(pkg._ff_install)
                path_remove(pkg._ff_post)
            elif (self.opts.pkg_action in
                    (PkgAction.RECONFIGURE, PkgAction.RECONFIGURE_ONLY)):
                path_remove(pkg._ff_bootstrap)
                path_remove(pkg._ff_configure)
                path_remove(pkg._ff_build)
                path_remove(pkg._ff_install)
                path_remove(pkg._ff_post)
            elif self.opts.pkg_action == PkgAction.REINSTALL:
                path_remove(pkg._ff_install)
                path_remove(pkg._ff_post)

        return True

    def _get_package_names(self, settings):
        """
        acquire list of project package names to process

        From a dictionary of user-defined settings, extract the known list of
        package names from the package configuration key ``ConfKey.PKGS``. This
        method will return a (duplicate-removed) list of packages (if any) to be
        processed.

        Args:
            settings: user settings to pull package information from

        Returns:
            list of package names to be processed; empty or ``None`` if no
            packages exist or an issue has occurred when attempting to extract
            package names
        """
        pkg_names = []
        bad_pkgs_value = False

        if ConfKey.PKGS in settings:
            pkg_names = interpret_strings(settings[ConfKey.PKGS])
            if pkg_names is None:
                bad_pkgs_value = True

        # remove duplicates (but maintain pre-sorted ordered)
        pkg_names = OrderedDict.fromkeys(pkg_names)

        if bad_pkgs_value:
            err('''\
bad package list definition

The configuration file does not have a properly formed list of defined packages.
Ensure a package list exists with the string-based names of packages to be part
of the releng process:

    {}
        {} = ['liba', 'libb', 'libc']''',
                self.opts.conf_point,
                ConfKey.PKGS)
        elif not pkg_names:
            raise RelengToolMissingPackagesError(
                self.opts.conf_point, ConfKey.PKGS)

        return pkg_names

    def _handle_clean_request(self, gaction):
        """
        handle a global clean request

        Performs a clean request for the working environment. A standard clean
        request will remove generated build, host, staging and target
        directories. In the event of "proper"-based cleans are requested,
        additional content such as the entire output directory (along with known
        file flags) can be removed.

        Args:
            gaction: the specific clean action being requested
        """

        if gaction == GlobalAction.DISTCLEAN:
            verbose('removing cache directory')
            path_remove(self.opts.cache_dir)
            verbose('removing download directory')
            path_remove(self.opts.dl_dir)

        if gaction in (GlobalAction.MRPROPER, GlobalAction.DISTCLEAN):
            verbose('removing output directory')
            path_remove(self.opts.out_dir)

            verbose('removing file flags')
            if os.path.exists(self.opts.ff_devmode):
                if path_remove(self.opts.ff_devmode):
                    warn('development mode has been unconfigured')
            if os.path.exists(self.opts.ff_local_srcs):
                if path_remove(self.opts.ff_local_srcs):
                    warn('local-sources mode has been unconfigured')
        else:
            verbose('removing build directory')
            path_remove(self.opts.build_dir)
            verbose('removing host directory')
            path_remove(self.opts.host_dir)
            verbose('removing license directory')
            path_remove(self.opts.license_dir)
            verbose('removing staging directory')
            path_remove(self.opts.staging_dir)
            verbose('removing symbols directory')
            path_remove(self.opts.symbols_dir)
            verbose('removing target directory')
            path_remove(self.opts.target_dir)

    def _perform_license_generation(self, license_files):
        """
        generate a license file for the project

        Compiles a document containing all the license information for a
        configured project. License information defined by each package will be
        populated into a single file after each package is examined. Returned
        will be a dictionary where keys are the package names and the values are
        the list of one or more license files for the package.

        Args:
            license_files: dictionary of each package's license files

        Returns:
            ``True`` if the license file was generated; ``False`` if the license
            file could not be generated
        """
        if not ensure_dir_exists(self.opts.license_dir):
            return False

        license_file = os.path.join(self.opts.license_dir, 'licenses')
        try:
            with open(license_file, 'w') as dst:
                license_header = expand(self.opts.license_header)
                if not license_header:
                    license_header = 'license(s)'

                # output license header
                dst.write('''{}
################################################################################
'''.format(license_header))

                # output license header
                for license_name, license_data in sorted(license_files.items()):
                    license_files = license_data['files']
                    license_version = license_data['version']
                    dst.write('''
{}-{}
--------------------------------------------------------------------------------
'''.format(license_name, license_version))
                    for pkg_license_file in sorted(license_files):
                        verbose('writing license file ({}): {}',
                            license_name, pkg_license_file)
                        with open(pkg_license_file, 'r') as f:
                            copyfileobj(f, dst)
                        dst.write('')

            verbose('license file has been written')
        except IOError as e:
            err('unable to populate license information\n'
                '    {}', e)
            return False

        return True

    def _post_processing(self, env):
        """
        perform post-processing of a release engineering process

        After processing all packages configured for a release engineering
        process, post-processing can be performed as the final stage (if the
        script exists). Typically, a build process will use this stage to
        package any generated results into a "final image".

        Args:
            env: environment settings to apply for processing script

        Returns:
            ``True`` if the post-processing script has executed correctly or no
            post-processing script exists; ``False`` if an error has occurred
            when processing the post-processing script
        """
        build_script, postbuild_exists = opt_file(self.opts.post_build_point)

        # TODO remove deprecated configuration load in (at maximum) v1.0
        if not postbuild_exists:
            deprecated_posts = [
                'releng-post',
                'post.py',
            ]

            for script_name in deprecated_posts:
                build_script = os.path.join(self.opts.root_dir, script_name)
                if os.path.isfile(build_script):
                    warn('using deprecated post-processing file {} -- switch '
                         'to {} for future projects', script_name,
                         RELENG_POST_BUILD_NAME)
                    postbuild_exists = True
                    break

        if postbuild_exists:
            verbose('performing post-processing (build)...')

            # ensure images directory exists (as the post-processing script will
            # most likely populate it)
            if not ensure_dir_exists(self.opts.images_dir):
                return False

            if not run_script(build_script, env, subject='post-build'):
                return False

            verbose('post-processing (build) completed')

        return True

    def _prepare_script_environment(self, script_env, gaction, paction):
        """
        prepare the script environment with common project values

        A package stage will be invoked with a tailored environment variables.
        This method is used to prepare an environment dictionary with common
        variables such the the staging directory, target directory and more.

        Args:
            script_env: environment dictionary to prepare
            gaction: the global action invoked (if any)
            paction: the package-specific action invoked (if any)
        """

        # always register optional flags in script environment
        script_env['RELENG_CLEAN'] = None
        script_env['RELENG_DEBUG'] = None
        script_env['RELENG_DEVMODE'] = None
        script_env['RELENG_DISTCLEAN'] = None
        script_env['RELENG_FORCE'] = None
        script_env['RELENG_LOCALSRCS'] = None
        script_env['RELENG_MRPROPER'] = None
        script_env['RELENG_REBUILD'] = None
        script_env['RELENG_RECONFIGURE'] = None
        script_env['RELENG_REINSTALL'] = None
        script_env['RELENG_VERBOSE'] = None

        #: default lib container directory
        sysroot_nprefix = os.path.normpath(self.opts.sysroot_prefix)
        host_pdir = self.opts.host_dir + sysroot_nprefix
        staging_pdir = self.opts.staging_dir + sysroot_nprefix
        target_pdir = self.opts.target_dir + sysroot_nprefix
        host_include_dir = os.path.join(host_pdir, 'include')
        host_lib_dir = os.path.join(host_pdir, 'lib')
        staging_include_dir = os.path.join(staging_pdir, 'include')
        staging_lib_dir = os.path.join(staging_pdir, 'lib')
        target_include_dir = os.path.join(target_pdir, 'include')
        target_lib_dir = os.path.join(target_pdir, 'lib')

        # global variables
        for env in (os.environ, script_env):
            env['BUILD_DIR'] = self.opts.build_dir
            env['CACHE_DIR'] = self.opts.cache_dir
            env['DL_DIR'] = self.opts.dl_dir
            env['HOST_DIR'] = self.opts.host_dir
            env['HOST_INCLUDE_DIR'] = host_include_dir
            env['HOST_LIB_DIR'] = host_lib_dir
            env['IMAGES_DIR'] = self.opts.images_dir
            env['LICENSE_DIR'] = self.opts.license_dir
            env['NJOBS'] = str(self.opts.jobs)
            env['NJOBSCONF'] = str(self.opts.jobsconf)
            env['OUTPUT_DIR'] = self.opts.out_dir
            env['PREFIX'] = self.opts.sysroot_prefix
            env['PREFIXED_HOST_DIR'] = host_pdir
            env['PREFIXED_STAGING_DIR'] = staging_pdir
            env['PREFIXED_TARGET_DIR'] = target_pdir
            env['RELENG_VERSION'] = releng_version
            env['ROOT_DIR'] = self.opts.root_dir
            env['STAGING_DIR'] = self.opts.staging_dir
            env['STAGING_INCLUDE_DIR'] = staging_include_dir
            env['STAGING_LIB_DIR'] = staging_lib_dir
            env['SYMBOLS_DIR'] = self.opts.symbols_dir
            env['TARGET_DIR'] = self.opts.target_dir
            env['TARGET_INCLUDE_DIR'] = target_include_dir
            env['TARGET_LIB_DIR'] = target_lib_dir

            if gaction == GlobalAction.CLEAN:
                env['RELENG_CLEAN'] = '1'
            elif gaction == GlobalAction.DISTCLEAN:
                env['RELENG_CLEAN'] = '1' # also set clean flag
                env['RELENG_DISTCLEAN'] = '1'
                env['RELENG_MRPROPER'] = '1' # also set mrproper flag
            elif gaction == GlobalAction.MRPROPER:
                env['RELENG_CLEAN'] = '1' # also set clean flag
                env['RELENG_MRPROPER'] = '1'

            if paction in (PkgAction.REBUILD, PkgAction.REBUILD_ONLY):
                env['RELENG_REBUILD'] = '1'
            elif paction in (PkgAction.RECONFIGURE, PkgAction.RECONFIGURE_ONLY):
                env['RELENG_RECONFIGURE'] = '1'
            elif paction == PkgAction.REINSTALL:
                env['RELENG_REINSTALL'] = '1'

            if self.opts.debug:
                env['RELENG_DEBUG'] = '1'
            if self.opts.devmode:
                env['RELENG_DEVMODE'] = '1'
            if self.opts.force:
                env['RELENG_FORCE'] = '1'
            if self.opts.local_srcs:
                env['RELENG_LOCALSRCS'] = '1'
            if self.opts.verbose:
                env['RELENG_VERBOSE'] = '1'

        # utility methods (if adjusting, see also `releng_tool.__init__`)
        script_env['debug'] = debug
        script_env['err'] = err
        script_env['log'] = log
        script_env['note'] = note
        script_env['releng_cat'] = cat
        script_env['releng_copy'] = path_copy
        script_env['releng_env'] = env_value
        script_env['releng_execute'] = execute
        script_env['releng_execute_rv'] = execute_rv
        script_env['releng_exists'] = path_exists
        script_env['releng_exit'] = platform_exit
        script_env['releng_expand'] = expand
        script_env['releng_include'] = releng_include
        script_env['releng_join'] = os.path.join
        script_env['releng_ls'] = ls
        script_env['releng_mkdir'] = ensure_dir_exists
        script_env['releng_move'] = path_move
        script_env['releng_remove'] = path_remove
        script_env['releng_require_version'] = require_version
        script_env['releng_tmpdir'] = generate_temp_dir
        script_env['releng_touch'] = touch
        script_env['releng_wd'] = interim_working_dir
        script_env['success'] = success
        script_env['verbose'] = verbose
        script_env['warn'] = warn

    def _process_file_flags(self):
        """
        check or configure known file flags

        For all known file flags, see if either respective flag options are set
        and/or configure file flags for which have been explicitly set to be
        enabled.

        Returns:
            ``None`` if the file flags has been processed and the options have
            been updated accordingly; ``True`` if the request to configure file
            flags has completed with no errors; ``False`` if the request to
            configure file flags failed to be performed
        """
        opts = self.opts
        configured = False
        err_flag = False

        state = process_file_flag(opts.devmode, opts.ff_devmode)
        if state == FileFlag.CONFIGURED:
            success('configured root for development mode')
            configured = True
        elif state == FileFlag.NOT_CONFIGURED:
            err_flag = True
        opts.devmode = (state == FileFlag.EXISTS)
        if opts.devmode:
            verbose('development mode enabled')

        state = process_file_flag(opts.local_srcs, opts.ff_local_srcs)
        if state == FileFlag.CONFIGURED:
            success('configured root for local-sources mode')
            configured = True
        elif state == FileFlag.NOT_CONFIGURED:
            err_flag = True
        opts.local_srcs = (state == FileFlag.EXISTS)
        if opts.local_srcs:
            verbose('local-sources mode enabled')

        if err_flag:
            return False
        elif configured:
            return True
        else:
            return None

    def _process_settings(self, settings):
        """
        process global settings provided from the user

        For all known file flags, see if either respective flag options are set
        and/or configure file flags for which have been explicitly set to be
        enabled.

        Args:
            settings: user settings to pull global information from

        Returns:
            ``None`` if the file flags has been processed and the options have
            been updated accordingly; ``True`` if the request to configure file
            flags has completed with no errors; ``False`` if the request to
            configure file flags failed to be performed
        """
        def notify_invalid_value(key, expected):
            err('''\
invalid configuration value provided

The configuration file defines a key with an unexpected type. Correct the
following key entry and re-try again.

    Key: {}
    Expected Type: {}''', key, expected)

        if ConfKey.CACHE_EXT_TRANSFORM in settings:
            cet = None
            if callable(settings[ConfKey.CACHE_EXT_TRANSFORM]):
                cet = settings[ConfKey.CACHE_EXT_TRANSFORM]
            if cet is None:
                notify_invalid_value(ConfKey.CACHE_EXT_TRANSFORM, 'callable')
                return False
            self.opts.cache_ext_transform = cet

        if ConfKey.DEFINTERN in settings:
            is_default_internal = settings[ConfKey.DEFINTERN]
            if not isinstance(is_default_internal, bool):
                notify_invalid_value(ConfKey.DEFINTERN, 'bool')
                return False
            self.opts.default_internal_pkgs = is_default_internal

        if ConfKey.LICENSE_HEADER in settings:
            license_header = interpret_string(settings[ConfKey.LICENSE_HEADER])
            if license_header is None:
                notify_invalid_value(ConfKey.LICENSE_HEADER, 'str')
                return False
            self.opts.license_header = license_header

        if ConfKey.OVERRIDE_REV in settings:
            orz = interpret_dictionary_strings(settings[ConfKey.OVERRIDE_REV])
            if orz is None:
                notify_invalid_value(ConfKey.OVERRIDE_REV, 'dict(str,str)')
                return False
            self.opts.revision_override = orz

        if ConfKey.OVERRIDE_SITES in settings:
            v = interpret_dictionary_strings(settings[ConfKey.OVERRIDE_SITES])
            if v is None:
                notify_invalid_value(ConfKey.OVERRIDE_SITES, 'dict(str,str)')
                return False
            self.opts.sites_override = v

        if ConfKey.OVERRIDE_TOOLS in settings:
            v = interpret_dictionary_strings(settings[ConfKey.OVERRIDE_TOOLS])
            if v is None:
                notify_invalid_value(ConfKey.OVERRIDE_TOOLS, 'dict(str,str)')
                return False
            self.opts.extract_override = v

        if ConfKey.PREREQUISITES in settings:
            prerequisites = interpret_strings(settings[ConfKey.PREREQUISITES])
            if prerequisites is None:
                notify_invalid_value(ConfKey.PREREQUISITES, 'str or list(str)')
                return False
            self.opts.prerequisites.extend(prerequisites)

        if ConfKey.QUIRKS in settings:
            quirks = interpret_strings(settings[ConfKey.QUIRKS])
            if quirks is None:
                notify_invalid_value(ConfKey.QUIRKS, 'str or list(str)')
                return False
            self.opts.quirks.extend(quirks)
            for quirk in quirks:
                verbose('configuration quirk applied: ' + quirk)

        if ConfKey.SYSROOT_PREFIX in settings:
            sysroot_prefix = interpret_string(settings[ConfKey.SYSROOT_PREFIX])
            if sysroot_prefix is None:
                notify_invalid_value(ConfKey.SYSROOT_PREFIX, 'str')
                return False
            if not sysroot_prefix.startswith('/'):
                sysroot_prefix = '/' + sysroot_prefix
            self.opts.sysroot_prefix = sysroot_prefix

        if ConfKey.URL_MIRROR in settings:
            url_mirror = interpret_string(settings[ConfKey.URL_MIRROR])
            if url_mirror is None:
                notify_invalid_value(ConfKey.URL_MIRROR, 'str')
                return False
            self.opts.url_mirror = url_mirror

        if ConfKey.URLOPEN_CONTEXT in settings:
            urlopen_context = None
            if isinstance(settings[ConfKey.URLOPEN_CONTEXT], ssl.SSLContext):
                urlopen_context = settings[ConfKey.URLOPEN_CONTEXT]
            if urlopen_context is None:
                notify_invalid_value(ConfKey.URLOPEN_CONTEXT, 'ssl.SSLContext')
                return False
            self.opts.urlopen_context = urlopen_context

        if ConfKey.EXTEN_PKGS in settings:
            epd = interpret_strings(settings[ConfKey.EXTEN_PKGS])
            if epd is None:
                notify_invalid_value(ConfKey.EXTEN_PKGS, 'str or list(str)')
                return False
            self.opts.extern_pkg_dirs = epd

        ext_names = []
        if ConfKey.EXTENSIONS in settings:
            ext_names = interpret_strings(settings[ConfKey.EXTENSIONS])
            if ext_names is None:
                notify_invalid_value(ConfKey.EXTENSIONS, 'str or list(str)')
                return False

        self.registry.load_all_extensions(ext_names)
        return True



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.defs import PackageType
from releng_tool.packages.exceptions import RelengToolInvalidPackageKeyValue
from releng_tool.packages.exceptions import RelengToolUnknownPackageType
from tests.support.pkg_config_test import TestPkgConfigsBase

class TestPkgConfigsPkgType(TestPkgConfigsBase):
    def test_pkgconfig_type_invalid_type(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('pkg-type-invalid-type')

    def test_pkgconfig_type_invalid_value(self):
        with self.assertRaises(RelengToolUnknownPackageType):
            self.LOAD('pkg-type-invalid-value')

    def test_pkgconfig_type_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertEqual(pkg.type, PackageType.SCRIPT)

    def test_pkgconfig_type_valid_autotools(self):
        pkg, _, _ = self.LOAD('pkg-type-valid-autotools')
        self.assertEqual(pkg.type, PackageType.AUTOTOOLS)

    def test_pkgconfig_type_valid_cmake(self):
        pkg, _, _ = self.LOAD('pkg-type-valid-cmake')
        self.assertEqual(pkg.type, PackageType.CMAKE)

    def test_pkgconfig_type_valid_python(self):
        pkg, _, _ = self.LOAD('pkg-type-valid-python')
        self.assertEqual(pkg.type, PackageType.PYTHON)

    def test_pkgconfig_type_valid_script(self):
        pkg, _, _ = self.LOAD('pkg-type-valid-script')
        self.assertEqual(pkg.type, PackageType.SCRIPT)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

# A flag (or option) can be driven by the existence of a file (i.e. a file
# flag). When an instance operates with file flags, two modes are approached in
# this implementation. Either the state of file flags are unknown and are
# attempted to be read. Once read, the instance can handle their process
# accordingly based off these flag states. The other path is if there is a
# request to configure file flags. If a file flag is being configured, the
# intent would be to configure the one (or multiple) file flag state and have
# the running instance shutdown.

from releng_tool.util.enum import Enum
from releng_tool.util.io import touch
from releng_tool.util.log import err
import os

class FileFlag(Enum):
    """
    file flag result states

    Attributes:
        CONFIGURED: file flag was configured
        EXISTS: file flag exists
        NOT_CONFIGURED: unable to configure the file flag
        NO_EXIST: file flag does not exist
    """
    CONFIGURED = 'configured'
    EXISTS = 'exists'
    NOT_CONFIGURED = 'not_configured'
    NO_EXIST = 'no_exist'

def check_file_flag(file):
    """
    check a file flag

    Attempt to read a file flag state by checking for the file's existence.

    Args:
        file: the filename

    Returns:
        ``FileFlag.EXISTS`` if the flag is enabled; ``FileFlag.NO_EXIST`` if the
            flag is not enabled
    """
    return process_file_flag(None, file)

def process_file_flag(flag, file, quiet=False):
    """
    process a file flag event

    Will either write a file flag configuration event or attempt to read a file
    flag state. If the ``flag`` option is set to ``True``, this process event
    will assume that this instance is attempting to configure a file flag (on)
    state and generate the target file flag on the system. If the flag option is
    set to ``False``, the file's existence will be checked to reflect whether or
    not the flag is considered enabled.

    Args:
        flag: the flag option to used; ``None`` to check flag state
        file: the filename
        quiet: suppression of any error messages to standard out

    Returns:
        ``FileFlag.EXISTS`` if the flag is enabled; ``FileFlag.NO_EXIST`` if the
            flag is not enabled; ``FileFlag.CONFIGURED`` if the flag was
            configured as requested; ``FileFlag.NOT_CONFIGURED`` if the flag
            could not be configured as requested
    """

    if flag:
        # When checking if the file flag exists, attempt to update the access/
        # modified times. For the case where may experience issues creating the
        # file flag themselves (permission errors, etc.), fallback on just the
        # existence of the file flag to still be considered as configured.
        if touch(file):
            rv = FileFlag.CONFIGURED
        else:
            if os.path.isfile(file):
                rv = FileFlag.CONFIGURED
            else:
                rv = FileFlag.NOT_CONFIGURED
                if not quiet:
                    err('unable to configure file flag: {}', file)
    elif flag is None and os.path.isfile(file):
        rv = FileFlag.EXISTS
    else:
        rv = FileFlag.NO_EXIST

    return rv



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

import os
import types

def extend_script_env(env, extra):
    """
    extend a partially filtered environment (globals) for a run_path event

    When invoking ``run_path`` [1], a dictionary of globals is provided to
    pre-populate a script's globals before execution. Inside the releng process,
    the command ``run_path`` is invoked several times to help load settings and
    package-specific scripts. To exist in sharing releng-provided constants and
    also assisting in allow some-level of sharing user-defined constants, the
    list of globals to be populated can be extended each execution and be passed
    into a following script. Not all global options are desired to be passed.
    For example, Python magic options and referenced built-in functions. This
    method can be used to easily extend an existing dictionary of globals while
    also filtering out undesired entries output from external scripts.

    [1]: https://docs.python.org/3/library/runpy.html

    Args:
        env: the environment to update
        extra: the globals to add to the environment

    Returns:
        the same environment passed in
    """
    extra_copy = extra.copy()

    for key, value in extra.items():
        # remove python magic objects (if any)
        if key.startswith('__') and key.endswith('__'):
            extra_copy.pop(key)
        # remove imported built-in functions
        elif isinstance(value, types.BuiltinFunctionType):
            extra_copy.pop(key)
        # remove imported functions
        elif isinstance(value, types.FunctionType):
            extra_copy.pop(key)
        # remove imported modules
        elif isinstance(value, types.ModuleType):
            extra_copy.pop(key)

    env.update(extra_copy)
    return env

# unique default helper for env_value
__ENV_VALUE_DEFAULT = object()

def env_value(key, value=__ENV_VALUE_DEFAULT):
    """
    helper to easily fetch or configure an environment variable

    Provides a caller a simple method to fetch or configure an environment
    variable for the current context. This call is the same as if one directly
    fetched from or managed a key-value with ``os.environ``. If ``value`` is not
    provided, the environment variable's value (if set) will be returned. If
    ``value`` is set to a value of ``None``, any set environment variable will
    be removed.

    An example when using in the context of script helpers is as follows:

    .. code-block:: python

        # get an environment variable
        value = releng_env('KEY')

        # set an environment variable
        releng_env('KEY', 'VALUE')

    Args:
        key: the environment key
        value (optional): the environment value to set

    Returns:
        the value of the environment variable
    """
    if value is __ENV_VALUE_DEFAULT:
        return os.environ.get(key)

    if value is None:
        if key in os.environ:
            del os.environ[key]
    else:
        os.environ[key] = value

    return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from distutils.command.clean import clean
from distutils import dir_util
from setuptools import find_packages
from setuptools import setup
import os

# remove extra resources not removed by the default clean operation
class ExtendedClean(clean):
    def run(self):
        clean.run(self)

        if not self.all:
            return

        extras = [
            'dist',
            'releng_tool.egg-info',
        ]
        for extra in extras:
            if os.path.exists(extra):
                dir_util.remove_tree(extra, dry_run=self.dry_run)

setup(
    cmdclass={
        'clean': ExtendedClean,
    },
    entry_points={
        'console_scripts': [
            'releng-tool = releng_tool.__main__:main',
        ],
    },
    extras_require={
        'statistics': [
            'matplotlib',
        ],
    },
    packages=find_packages(exclude=["tests*"]),
    test_suite='tests',
    version='0.12.0',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool.api import RelengFetchOptions
from releng_tool.defs import VcsType
from releng_tool.fetch.bzr import fetch as fetch_bzr
from releng_tool.fetch.cvs import fetch as fetch_cvs
from releng_tool.fetch.git import fetch as fetch_git
from releng_tool.fetch.mercurial import fetch as fetch_mercurial
from releng_tool.fetch.rsync import fetch as fetch_rsync
from releng_tool.fetch.scp import fetch as fetch_scp
from releng_tool.fetch.svn import fetch as fetch_svn
from releng_tool.fetch.url import fetch as fetch_url
from releng_tool.tool.gpg import GPG
from releng_tool.util.api import replicate_package_attribs
from releng_tool.util.hash import HashResult
from releng_tool.util.hash import verify as verify_hashes
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.io import generate_temp_dir
from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import path_remove
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import verbose
from releng_tool.util.log import warn
import os
import shutil

def stage(engine, pkg, ignore_cache, extra_opts):
    """
    handles the fetching stage for a package

    With a provided engine and package instance, the fetching stage will be
    processed.

    Args:
        engine: the engine
        pkg: the package being fetched
        ignore_cache: always attempt to ignore the cache
        extra_opts: extra options for the fetch operation (if applicable)

    Returns:
        ``True`` if the fetching stage is completed; ``False`` otherwise
    """
    assert pkg.vcs_type
    name = pkg.name
    debug('process fetch stage: ' + name)

    # local sources mode requires internal sources to be already checked out
    if pkg.is_internal and engine.opts.local_srcs:
        if os.path.isdir(pkg.build_dir):
            return True

        err('''\
missing local sources for internal package: {0}

The active configuration is flagged for 'local sources' mode; however, an
internal package cannot be found in the local system. Before continuing, ensure
you have checked out all internal packages on your local system (or, disable the
local sources option to use the default process).

       Package: {0}
 Expected Path: {1}''', name, pkg.build_dir)
        return False

    # if the vcs-type is archive-based, flag that hash checks are needed
    perform_file_asc_check = False
    perform_file_hash_check = False
    if pkg.vcs_type == VcsType.URL:
        perform_file_asc_check = os.path.exists(pkg.asc_file)
        perform_file_hash_check = True

    fetch_opts = RelengFetchOptions()
    replicate_package_attribs(fetch_opts, pkg)
    fetch_opts.cache_dir = pkg.cache_dir
    fetch_opts.ext = pkg.ext_modifiers
    fetch_opts.extra_opts = extra_opts
    fetch_opts.ignore_cache = ignore_cache
    fetch_opts.name = name
    fetch_opts.revision = pkg.revision
    fetch_opts.site = pkg.site
    fetch_opts.version = pkg.version
    fetch_opts._mirror = False
    fetch_opts._quirks = engine.opts.quirks
    fetch_opts._urlopen_context = engine.opts.urlopen_context

    cache_filename = os.path.basename(pkg.cache_file)
    out_dir = engine.opts.out_dir
    with generate_temp_dir(out_dir) as work_dir, \
            generate_temp_dir(out_dir) as interim_cache_dir:
        with interim_working_dir(work_dir):
            interim_cache_file = os.path.join(interim_cache_dir, cache_filename)
            fetch_opts.cache_file = interim_cache_file
            fetch_opts.work_dir = work_dir

            # check if file caching should be ignored
            #
            # In special cases, a developer may configure a project to have a
            # fetched source not to cache. For example, pulling from a branch of
            # a VCS source will make a cache file from the branch and will
            # remain until manually removed from a cache file. A user may wish
            # to re-build the local cache file after cleaning their project.
            # While the releng-tool framework separates fetching/extraction into
            # two parts, ignoring cached assets can be partially achieved by
            # just removing any detected cache file if a project is configured
            # to ignore a cache file.
            if engine.opts.devmode and pkg.devmode_ignore_cache is not None:
                fetch_opts.ignore_cache = pkg.devmode_ignore_cache

                if pkg.devmode_ignore_cache and os.path.exists(pkg.cache_file):
                    verbose('removing cache file (per configuration): ' + name)
                    if not path_remove(pkg.cache_file):
                        return False

            # remove cache file if there is a force request to ignore the cache
            elif engine.opts.force and ignore_cache:
                if os.path.exists(pkg.cache_file):
                    verbose('removing cache file (forced): ' + name)
                    if not path_remove(pkg.cache_file):
                        return False

            # force explicit ignore cache (to off) when not in development mode
            elif not engine.opts.devmode and ignore_cache is None:
                fetch_opts.ignore_cache = False

            if os.path.exists(pkg.cache_file):
                rv = None
                if perform_file_hash_check:
                    hr = verify_hashes(
                        pkg.hash_file, pkg.cache_file, relaxed=True)

                    if hr == HashResult.VERIFIED:
                        rv = True
                    elif hr == HashResult.BAD_PATH:
                        if not perform_file_asc_check and not pkg.is_internal:
                            warn('missing hash file for package: ' + name)
                        rv = True # no hash file to compare with; assuming ok
                    elif hr == HashResult.EMPTY:
                        if not pkg.is_internal:
                            warn('hash file for package is empty: ' + name)
                        rv = True # empty hash file; assuming ok
                    elif hr == HashResult.MISMATCH:
                        if not path_remove(pkg.cache_file):
                            rv = False
                    elif hr in (HashResult.BAD_FORMAT, HashResult.UNSUPPORTED):
                        rv = False
                    elif hr == HashResult.MISSING_ARCHIVE:
                        if not perform_file_asc_check:
                            err('''\
missing archive hash for verification

The hash file for this package does not have an entry for the cache file to be
verified. Ensure the hash file defines an entry for the expected cache file:

    Hash File: {}
         File: {}''', pkg.hash_file, cache_filename)
                            rv = False
                    else:
                        err('invalid fetch operation (internal error; '
                            'hash-check failure: {})', hr)
                        rv = False
                else:
                    rv = True

                if rv is not False and perform_file_asc_check and \
                        os.path.exists(pkg.cache_file):
                    if GPG.validate(pkg.asc_file, pkg.cache_file):
                        rv = True
                    else:
                        if not path_remove(pkg.cache_file):
                            err('''\
failed to validate against ascii-armor

Validation of a package resource failed to verify against a provided ASCII-armor
file. Ensure that the package's public key has been registered into gpg.

 ASC File: {}
     File: {}''', pkg.asc_file, cache_filename)
                            rv = False
                        else:
                            rv = None

                if rv is not None:
                    if ignore_cache:
                        verbose('ignoring cache not supported for package: {}',
                            name)
                    return rv

            # find fetching method for the target vcs-type
            fetcher = None
            if pkg.vcs_type in engine.registry.fetch_types:
                def _(opts):
                    return engine.registry.fetch_types[pkg.vcs_type].fetch(
                        pkg.vcs_type, opts)
                fetcher = _
            elif pkg.vcs_type == VcsType.BZR:
                fetcher = fetch_bzr
            elif pkg.vcs_type == VcsType.CVS:
                fetcher = fetch_cvs
            elif pkg.vcs_type == VcsType.GIT:
                fetcher = fetch_git
            elif pkg.vcs_type == VcsType.HG:
                fetcher = fetch_mercurial
            elif pkg.vcs_type == VcsType.RSYNC:
                fetcher = fetch_rsync
            elif pkg.vcs_type == VcsType.SCP:
                fetcher = fetch_scp
            elif pkg.vcs_type == VcsType.SVN:
                fetcher = fetch_svn
            elif pkg.vcs_type == VcsType.URL:
                fetcher = fetch_url

            if not fetcher:
                err('fetch type is not implemented: {}', pkg.vcs_type)
                return False

            # if this is url-type location, attempt to search on the mirror
            # first (if configured)
            fetched = None
            if engine.opts.url_mirror and pkg.vcs_type == VcsType.URL:
                original_site = fetch_opts.site
                new_site = engine.opts.url_mirror + cache_filename
                if original_site != new_site:
                    fetch_opts._mirror = True

                    fetch_opts.site = new_site
                    fetched = fetcher(fetch_opts)
                    fetch_opts.site = original_site

                    fetch_opts._mirror = False

            # perform the fetch request (if not already fetched)
            if not fetched:
                fetched = fetcher(fetch_opts)
                if not fetched:
                    return False

            # if the fetch type has populated the package's cache directory
            # directly, we are done
            if fetched == pkg.cache_dir:
                pass
            # if the fetch type has returned a file, the file needs to be hash
            # checked and then be moved into the download cache
            elif fetched == interim_cache_file:
                if perform_file_hash_check:
                    hr = verify_hashes(pkg.hash_file, fetched)
                    if hr == HashResult.VERIFIED:
                        pass
                    elif hr == HashResult.BAD_PATH:
                        if not perform_file_asc_check and not pkg.is_internal:
                            warn('missing hash file for package: ' + name)
                    elif hr == HashResult.EMPTY:
                        if not pkg.is_internal:
                            warn('hash file for package is empty: ' + name)
                    elif hr == HashResult.MISMATCH:
                        return False
                    elif hr in (HashResult.BAD_FORMAT, HashResult.UNSUPPORTED):
                        return False
                    elif hr == HashResult.MISSING_ARCHIVE:
                        if not perform_file_asc_check:
                            err('''\
missing archive hash for verification

The hash file for this package does not have an entry for the cache file to be
verified. Ensure the hash file defines an entry for the expected cache file:

    Hash File: {}
         File: {}''', pkg.hash_file, cache_filename)
                            return False
                    else:
                        err('invalid fetch operation (internal error; '
                            'hash-check failure: {})', hr)
                        return False

                if perform_file_asc_check:
                    if not GPG.validate(pkg.asc_file, interim_cache_file):
                        err('''\
failed to validate against ascii-armor

Validation of a package resource failed to verify against a provided ASCII-armor
file. Ensure that the package's public key has been registered into gpg.

     ASC File: {}
         File: {}''', pkg.asc_file, cache_filename)
                        return False

                debug('fetch successful; moving cache file')

                # ensure the download directory exists
                if not ensure_dir_exists(engine.opts.dl_dir):
                    return False

                try:
                    shutil.move(interim_cache_file, pkg.cache_file)
                except shutil.Error:
                    err('invalid fetch operation (internal error; fetch mode '
                        '"{}" has provided a missing cache file)', pkg.vcs_type)
                    return False
            else:
                err('invalid fetch operation (internal error; fetch mode "{}" '
                    'has returned an unsupported value)', pkg.vcs_type)
                return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.make import MAKE
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.string import expand

def build(opts):
    """
    support building autotools projects

    With provided build options (``RelengBuildOptions``), the build stage will
    be processed.

    Args:
        opts: build options

    Returns:
        ``True`` if the building stage is completed; ``False`` otherwise
    """

    if not MAKE.exists():
        err('unable to build package; make is not installed')
        return False

    # definitions
    autotools_defs = {
    }
    if opts.build_defs:
        autotools_defs.update(expand(opts.build_defs))

    # default options
    autotools_opts = {
    }
    if opts.build_opts:
        autotools_opts.update(expand(opts.build_opts))

    # argument building
    autotools_args = [
    ]
    autotools_args.extend(prepare_definitions(autotools_defs))
    autotools_args.extend(prepare_arguments(autotools_opts))

    if opts.jobs > 1:
        autotools_args.append('--jobs')
        autotools_args.append(str(opts.jobs))

    if not MAKE.execute(autotools_args, env=expand(opts.build_env)):
        err('failed to build autotools project: {}', opts.name)
        return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2020 releng-tool
# flake8: noqa

from releng_tool.api import *

if '_RELENG_GLOBAL_NAMESPACE_API_DEPRECATED' not in globals():
    global _RELENG_GLOBAL_NAMESPACE_API_DEPRECATED
    _RELENG_GLOBAL_NAMESPACE_API_DEPRECATED = True
    from releng_tool.util.log import warn as __warn
    __warn('(deprecated) namespace releng.api has been deprecated over releng_tool.api')



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.tool.rsync import RSYNC
from releng_tool.util.io import interpret_stem_extension
from releng_tool.util.io import prepare_arguments
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import note
from releng_tool.util.string import expand
import os
import sys
import tarfile

def fetch(opts):
    """
    support fetching from rsync sources

    With provided fetch options (``RelengFetchOptions``), the fetch stage will
    be processed.

    Args:
        opts: fetch options

    Returns:
        ``True`` if the fetch stage is completed; ``False`` otherwise
    """

    assert opts
    cache_file = opts.cache_file
    name = opts.name
    site = opts.site
    work_dir = opts.work_dir

    cache_basename = os.path.basename(cache_file)
    cache_stem, __ = interpret_stem_extension(cache_basename)

    if not RSYNC.exists():
        err('unable to fetch package; rsync is not installed')
        return None

    note('fetching {}...', name)
    sys.stdout.flush()

    # options
    fetch_opts = {
        '--recursive': '',  # default recursive call
    }
    if opts.extra_opts:
        fetch_opts.update(expand(opts.extra_opts))

    # argument building
    fetch_args = [
    ]
    fetch_args.extend(prepare_arguments(fetch_opts))

    # sanity check provided arguments
    for fetch_arg in fetch_args:
        if '--remove-source-files' in fetch_arg:
            err('option `--remove-source-files` not permitted')
            return None
        elif not fetch_arg.startswith('-'):
            err('invalid fetch option provided:', fetch_arg)
            return None

    fetch_args.append(site)  # source directory
    fetch_args.append(work_dir)  # destination directory

    if not RSYNC.execute(fetch_args, cwd=work_dir):
        err('unable to rsync from source')
        return None
    log('successfully invoked rsync for source')

    with tarfile.open(cache_file, 'w:gz') as tar:
        tar.add(work_dir, arcname=cache_stem)

    return cache_file



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.api import RelengExtractExtensionInterface
from releng_tool.opts import RelengEngineOptions
from releng_tool.packages.exceptions import RelengToolInvalidPackageKeyValue
from releng_tool.packages.exceptions import RelengToolUnknownExtractType
from releng_tool.packages.manager import RelengPackageManager
from releng_tool.registry import RelengRegistry
from tests.support.pkg_config_test import TestPkgConfigsBase

class TestPkgConfigsExtensions(TestPkgConfigsBase):
    def test_pkgconfig_ext_mods_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('ext-mods-invalid-type')

    def test_pkgconfig_ext_mods_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.ext_modifiers)

    def test_pkgconfig_ext_mods_valid(self):
        pkg, _, _ = self.LOAD('ext-mods-valid-empty')
        self.assertDictEqual(pkg.ext_modifiers, {})

        pkg, _, _ = self.LOAD('ext-mods-valid-data')
        self.assertDictEqual(pkg.ext_modifiers, {
            'mod1': 'value',
            'mod2': True,
            123: {
                'more': 'data',
            },
        })

    def test_pkgconfig_extract_type_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('extract-type-invalid-type')

    def test_pkgconfig_extract_type_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.extract_type)

    def test_pkgconfig_extract_type_registered(self):
        opts = RelengEngineOptions()
        registry = RelengRegistry()
        manager = RelengPackageManager(opts, registry)

        # register the custom extraction type
        class MockExtract(RelengExtractExtensionInterface):
            def extract(self, name, opts):
                return False

        CUSTOM_EXTRACT_NAME = 'ext-my-custom-extract-type'
        registry.add_extract_type(CUSTOM_EXTRACT_NAME, MockExtract)

        pkg, _, _ = self.LOAD('extract-type-valid', manager=manager)
        self.assertEqual(pkg.extract_type, CUSTOM_EXTRACT_NAME)

    def test_pkgconfig_extract_type_unregistered(self):
        with self.assertRaises(RelengToolUnknownExtractType):
            self.LOAD('extract-type-valid')



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.opts import RelengEngineOptions
from releng_tool.packages.exceptions import RelengToolInvalidPackageKeyValue
from releng_tool.packages.manager import RelengPackageManager
from releng_tool.registry import RelengRegistry
from tests.support.pkg_config_test import TestPkgConfigsBase

class TestPkgConfigsDevmode(TestPkgConfigsBase):
    def test_pkgconfig_devmode_ignore_cache_disabled(self):
        pkg, _, _ = self.LOAD('devmode-ignore-cache-disabled')
        self.assertEqual(pkg.devmode_ignore_cache, False)

    def test_pkgconfig_devmode_ignore_cache_enabled(self):
        pkg, _, _ = self.LOAD('devmode-ignore-cache-enabled')
        self.assertEqual(pkg.devmode_ignore_cache, True)

    def test_pkgconfig_devmode_ignore_cache_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('devmode-ignore-cache-invalid')

    def test_pkgconfig_devmode_ignore_cache_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.devmode_ignore_cache)

    def test_pkgconfig_devmode_revision_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('devmode-revision-invalid-type')

    def test_pkgconfig_devmode_revision_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertFalse(pkg.has_devmode_option)

    def test_pkgconfig_devmode_revision_valid_default(self):
        pkg, _, _ = self.LOAD('devmode-revision-valid')
        self.assertEqual(pkg.revision, 'dummy')
        self.assertEqual(pkg.version, 'dummy')
        self.assertTrue(pkg.has_devmode_option)

    def test_pkgconfig_devmode_revision_valid_enabled(self):
        # force engine options to default packages to internal
        opts = RelengEngineOptions()
        opts.devmode = True

        registry = RelengRegistry()
        manager = RelengPackageManager(opts, registry)

        pkg, _, _ = self.LOAD('devmode-revision-valid', manager=manager)
        self.assertEqual(pkg.revision, 'my-devmode-revision')
        self.assertEqual(pkg.version, 'my-devmode-revision')
        self.assertTrue(pkg.has_devmode_option)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

# sample extension will should load since the the required version
# number is older than the current version
def releng_setup(app):
    app.require_version('0.0')



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from copy import deepcopy

def replicate_package_attribs(opts, pkg):
    """
    replicate package attributes into an engine options entity

    Process a package entity (``RelengPackage``) and copies public attributes of
    the class into a provided options entity. This is to help expose internally
    tracked package options into internally managed engine stages (e.g. allowing
    the script build process to use all package options not defined in the API).

    Args:
        opts: the options to update
        pkg: the package entity with attributes to replicate
    """
    for key, value in pkg.__dict__.items():
        if not key.startswith('_'):
            opts.__dict__['_' + key] = deepcopy(value)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from __future__ import print_function
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import note
from releng_tool.util.log import warn
import contextlib
import os
import sys

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

#: size of blocks read when downloading a resource
REQUEST_READ_BLOCKSIZE = 8192

def fetch(opts):
    """
    support fetching from url sources

    With provided fetch options (``RelengFetchOptions``), the fetch stage will
    be processed.

    Args:
        opts: fetch options

    Returns:
        ``True`` if the fetch stage is completed; ``False`` otherwise
    """

    assert opts
    cache_file = opts.cache_file
    name = opts.name
    site = opts.site
    is_mirror_attempt = opts._mirror
    urlopen_context = opts._urlopen_context

    filename = os.path.basename(cache_file)

    note('fetching {}...', name)
    sys.stdout.flush()

    log('requesting: ' + site)
    try:
        with contextlib.closing(urlopen(site, context=urlopen_context)) as rsp:
            total = 0
            if 'content-length' in rsp.headers:
                try:
                    total = int(rsp.headers['content-length'])
                    total_str = display_size(total)
                except ValueError:
                    pass

            read = 0
            with open(cache_file, 'wb') as f:
                while True:
                    buf = rsp.read(REQUEST_READ_BLOCKSIZE)
                    if not buf:
                        break
                    read += len(buf)
                    read_str = display_size(read)

                    if total != read:
                        if total > 0:
                            pct = 100 * float(read) / float(total)
                            print('[{:02.0f}%] {}: {} of {}            '.format(
                                pct, filename, read_str, total_str), end='\r')
                        else:
                            print(' {}: {}            '.format(
                                filename, read_str), end='\r')

                    f.write(buf)
    except Exception as e:
        log_func = warn if is_mirror_attempt else err
        log_func('failed to download resource\n'
            '    {}', e)
        return None

    # cleanup any download progress prints
    if read > 0:
        log('')

    log('completed download ({})', display_size(read))
    return cache_file

def display_size(val):
    """
    return a human-readable count value for the provided byte count

    Accepts a byte count value and returns a string with a count value and
    binary prefix which describes the respective size.

    Args:
        val: the value (in bytes) to interpret

    Returns:
        the human-readable size
    """
    for unit in ['B', 'KiB', 'MiB', 'GiB']:
        if abs(val) < 1024.:
            return '{:3.1f} {}'.format(val, unit)
        val /= 1024.

    return '{:.1f} TiB'.format(val)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool
from releng_tool.util.io import execute_rv
from releng_tool.util.log import debug

#: executable used to run tar commands
TAR_COMMAND = 'tar'

#: list of environment keys to filter from a environment dictionary
TAR_SANITIZE_ENV_KEYS = [
    'TAR_OPTIONS',
]

class TarTool(RelengTool):
    """
    tar host tool

    Provides addition helper methods for tar-based tool interaction.

    Attributes:
        force_local: whether or not the `force-local` option is supported
    """

    force_local = False

    def exists(self):
        """
        return whether or not the host tool exists

        Returns whether or not the tool is available on the host for use.

        Returns:
            ``True``, if the tool exists; ``False`` otherwise
        """
        if self.tool in RelengTool.detected:
            return RelengTool.detected[self.tool]

        rv, out = execute_rv(self.tool, '--help')
        if rv == 0:
            debug('{} tool is detected on this system', self.tool)
            if '--force-local' in out:
                debug('{} tool supports force-local', self.tool)
                TarTool.force_local = True
            else:
                debug('{} tool does not support force-local', self.tool)
            RelengTool.detected[self.tool] = True
        else:
            debug('{} tool is not detected on this system', self.tool)
            RelengTool.detected[self.tool] = False

        return RelengTool.detected[self.tool]

#: tar host tool helper
TAR = TarTool(TAR_COMMAND, env_sanitize=TAR_SANITIZE_ENV_KEYS)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from tests import find_test_base
import os

def fetch_unittest_assets_dir(*args):
    """
    fetch the unit tests assets directory

    Will return the full path for the unit testing's asset directory.

    Args:
        *args (optional): path entries to append to the asset directory

    Returns:
        the directory
    """
    base_dir = find_test_base()
    return os.path.join(base_dir, 'unit-tests', 'assets', *args)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from tests import prepare_testenv
import os
import unittest

class TestPkgPipelineRemoteScripts(unittest.TestCase):
    def test_pkg_pipeline_remote_scripts_disabled_option(self):
        with prepare_testenv(template='remote-scripts-disabled') as engine:
            engine.run()
            self._assertFileFlag(engine, 'build-remote', False)
            self._assertFileFlag(engine, 'configure-remote', False)
            self._assertFileFlag(engine, 'install-remote', False)

    def test_pkg_pipeline_remote_scripts_disabled_quirk(self):
        conf = {
            'quirk': [
                'releng.disable_remote_scripts',
            ],
        }
        with prepare_testenv(config=conf, template='remote-scripts') as engine:
            engine.run()
            self._assertFileFlag(engine, 'build-remote', False)
            self._assertFileFlag(engine, 'configure-remote', False)
            self._assertFileFlag(engine, 'install-remote', False)

    def test_pkg_pipeline_remote_scripts_enabled(self):
        with prepare_testenv(template='remote-scripts') as engine:
            engine.run()
            self._assertFileFlag(engine, 'build-remote', True)
            self._assertFileFlag(engine, 'configure-remote', True)
            self._assertFileFlag(engine, 'install-remote', True)

    def test_pkg_pipeline_remote_scripts_override_all(self):
        with prepare_testenv(template='remote-scripts-override-all') as engine:
            engine.run()
            self._assertFileFlag(engine, 'build-override', True)
            self._assertFileFlag(engine, 'build-remote', False)
            self._assertFileFlag(engine, 'configure-override', True)
            self._assertFileFlag(engine, 'configure-remote', False)
            self._assertFileFlag(engine, 'install-override', True)
            self._assertFileFlag(engine, 'install-remote', False)

    def test_pkg_pipeline_remote_scripts_override_subset(self):
        template = 'remote-scripts-override-subset'
        with prepare_testenv(template=template) as engine:
            engine.run()
            self._assertFileFlag(engine, 'build-remote', True)
            self._assertFileFlag(engine, 'configure-remote', True)
            self._assertFileFlag(engine, 'install-override', True)
            self._assertFileFlag(engine, 'install-remote', False)

    def _assertFileFlag(self, engine, name, exists):
        file_flag = os.path.join(engine.opts.target_dir, name)
        self.assertEqual(os.path.exists(file_flag), exists)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.cmake import CMAKE
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.log import verbose
from releng_tool.util.string import expand

def install(opts):
    """
    support installation cmake projects

    With provided installation options (``RelengInstallOptions``), the
    installation stage will be processed.

    Args:
        opts: installation options

    Returns:
        ``True`` if the installation stage is completed; ``False`` otherwise
    """

    if not CMAKE.exists():
        err('unable to install package; cmake is not installed')
        return False

    # check if the no-install flag is set
    if opts._cmake_noinstall:
        verbose('configured to skip install stage for cmake')
        return True

    # default definitions
    cmake_defs = {
    }
    if opts.install_defs:
        cmake_defs.update(expand(opts.install_defs))

    # default options
    cmake_opts = {
        # build RelWithDebInfo (when using multi-configuration projects)
        '--config': 'RelWithDebInfo',
        # default install using the install target
        '--target': 'install',
    }
    if opts.install_opts:
        cmake_opts.update(expand(opts.install_opts))

    # argument building
    cmake_args = [
        '--build',
        opts.build_output_dir,
    ]
    cmake_args.extend(prepare_definitions(cmake_defs, '-D'))
    cmake_args.extend(prepare_arguments(cmake_opts))

    # prepare environment for installation request; an environment dictionary is
    # always needed to apply a custom DESTDIR during each install request
    env = expand(opts.install_env)
    if not env:
        env = {}

    # install to each destination
    for dest_dir in opts.dest_dirs:
        env['DESTDIR'] = dest_dir
        if not CMAKE.execute(cmake_args, env=env):
            err('failed to install cmake project: {}', opts.name)
            return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool

#: executable used to run autoreconf commands
AUTORECONF_COMMAND = 'autoreconf'

#: autoreconf host tool helper
AUTORECONF = RelengTool(AUTORECONF_COMMAND)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

class RelengPackage:
    """
    a releng package

    A package tracks the name, options and dependencies of the package.

    Args:
        name: the name of the package
        version: the version of the package

    Attributes:
        asc_file: file containing ascii-armored data to validate this package
        build_dir: directory for a package's buildable content
        build_output_dir: build output directory for the package process
        build_subdir: override for a package's buildable content (if applicable)
        build_tree: the build tree directory for a package
        cache_dir: cache directory for the package (if applicable)
        cache_file: cache file for the package (if applicable)
        def_dir: directory for the package definition
        deps: list of dependencies for this package
        devmode_ignore_cache: whether or not cache files should be ignored
        ext_modifiers: extension-defined modifiers (dict)
        extract_type: extraction type override (for extensions, if applicable)
        fetch_opts: fetch options (if applicable)
        fixed_jobs: fixed job count for this specific package
        git_config: git config options to apply (if applicable)
        git_depth: git fetch depth (if applicable)
        git_refspecs: additional git refspecs to fetch (if applicable)
        git_submodules: fetch any git submodules (if applicable)
        git_verify_revision: verify signed git revisions
        has_devmode_option: whether or not the package has a devmode revision
        hash_file: file containing hashes to validate this package
        install_type: install container for the package (target, staged, etc.)
        is_internal: whether or not this package is an project internal package
        license: license(s) of the package
        license_files: list of files in sources holding license information
        name: name of the package
        no_extraction: whether or not this package will extract
        nv: name-version value of the package
        prefix: system root prefix override (if applicable)
        revision: revision to use to fetch from vcs (if applicable)
        site: site to acquire package assets
        skip_remote_config: whether or not to skip any remote configuration
        skip_remote_scripts: whether or not to skip any remote scripts
        strip_count: archive extraction strip count (if applicable)
        type: package type (script-based, cmake, etc.)
        vcs_type: vcs type of the package (git, file, etc.)
        version: package version
        (package type - common)
        build_defs: package-type build definitions
        build_env: package-type build environment overrides
        build_opts: package-type build option overrides
        conf_defs: package-type configuration definitions
        conf_env: package-type configuration environment overrides
        conf_opts: package-type configuration option overrides
        install_defs: package-type installation definitions
        install_env: package-type installation environment overrides
        install_opts: package-type installation option overrides
        (package type - autotools)
        autotools_autoreconf: flag to invoke autoreconf
        (package type - cmake)
        cmake_noinstall: flag to disable the install stage for a cmake project
        (other - python)
        python_interpreter: python interpreter to invoke stages with
    """
    def __init__(self, name, version):
        self.name = name
        self.version = version
        if version:
            self.nv = '{}-{}'.format(name, version)
        else:
            self.nv = self.name
        # (commons)
        self.asc_file = None
        self.build_dir = None
        self.build_subdir = None
        self.build_output_dir = None
        self.build_tree = None
        self.cache_dir = None
        self.cache_file = None
        self.def_dir = None
        self.deps = []
        self.devmode_ignore_cache = None
        self.fetch_opts = None
        self.fixed_jobs = None
        self.has_devmode_option = None
        self.hash_file = None
        self.ext_modifiers = None
        self.extract_type = None
        self.install_type = None
        self.is_internal = None
        self.license = None
        self.license_files = None
        self.no_extraction = False
        self.prefix = None
        self.revision = None
        self.site = None
        self.skip_remote_config = None
        self.skip_remote_scripts = None
        self.strip_count = None
        self.type = None
        self.vcs_type = None
        # (package type - common)
        self.build_defs = None
        self.build_env = None
        self.build_opts = None
        self.conf_defs = None
        self.conf_env = None
        self.conf_opts = None
        self.install_defs = None
        self.install_env = None
        self.install_opts = None
        # (package type - autotools)
        self.autotools_autoreconf = None
        # (package type - cmake)
        self.cmake_noinstall = None
        # (other - git)
        self.git_config = None
        self.git_depth = None
        self.git_refspecs = None
        self.git_submodules = None
        self.git_verify_revision = None
        # (other - python)
        self.python_interpreter = None



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool.defs import PackageInstallType
from releng_tool.tool.cmake import CMAKE
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.string import expand
import os
import posixpath

#: default lib container directory
DEFAULT_LIB_DIR = 'lib'

def configure(opts):
    """
    support configuration for cmake projects

    With provided configuration options (``RelengConfigureOptions``), the
    configuration stage will be processed.

    Args:
        opts: configuration options

    Returns:
        ``True`` if the configuration stage is completed; ``False`` otherwise
    """

    if not CMAKE.exists():
        err('unable to configure package; cmake is not installed')
        return False

    prefix = opts.prefix

    base_locs = []
    if opts._install_type == PackageInstallType.HOST:
        base_locs.append(opts.host_dir)
    else:
        base_locs.append(opts.staging_dir)
        base_locs.append(opts.target_dir)

    include_locs = []
    library_locs = []
    prefix_locs = []
    for base_loc in base_locs:
        prefixed_base = base_loc + prefix
        include_locs.append(os.path.join(prefixed_base, 'include'))
        library_locs.append(os.path.join(prefixed_base, DEFAULT_LIB_DIR))
        prefix_locs.append(prefixed_base)

    lib_dir = os.path.join(prefix, DEFAULT_LIB_DIR)

    # ensure the non-full prefix options are passed in a posix style, or
    # some versions of CMake/projects may treat the path separators as
    # escape characters
    posix_prefix = prefix.replace(os.sep, posixpath.sep)
    posix_lib_dir = lib_dir.replace(os.sep, posixpath.sep)

    # definitions
    cmake_defs = {
        # configure as RelWithDebInfo (when using multi-configuration projects)
        'CMAKE_BUILD_TYPE': 'RelWithDebInfo',
        # common paths for releng-tool sysroots
        'CMAKE_INCLUDE_PATH': ';'.join(include_locs),
        'CMAKE_INSTALL_PREFIX': posix_prefix,
        'CMAKE_LIBRARY_PATH': ';'.join(library_locs),
        'CMAKE_PREFIX_PATH': ';'.join(prefix_locs),
        # releng-tool's sysroot assumes a `lib` directory. CMake's
        # GNUInstallDirs may adjust the expected lib directory based on the
        # detected system name (as a project may not necessarily be
        # cross-compiling), which may implicitly set the library directory to
        # `lib64`.
        'CMAKE_INSTALL_LIBDIR': posix_lib_dir,
    }
    if opts.conf_defs:
        cmake_defs.update(expand(opts.conf_defs))

    # options
    cmake_opts = {
    }
    if opts.conf_opts:
        cmake_opts.update(expand(opts.conf_opts))

    # argument building
    cmake_args = [
        '--no-warn-unused-cli', # suppress common options if not used in project
    ]
    cmake_args.extend(prepare_definitions(cmake_defs, '-D'))
    cmake_args.extend(prepare_arguments(cmake_opts))

    # output directory
    cmake_args.append(opts.build_dir)

    # ensure provided include/library targets exists ahead of time to help
    # reduce the risk of CMake projects creating files for these directory paths
    populate_dirs = []
    if 'CMAKE_INCLUDE_PATH' in cmake_defs:
        populate_dirs.extend(cmake_defs['CMAKE_INCLUDE_PATH'].split(';'))
    if 'CMAKE_LIBRARY_PATH' in cmake_defs:
        populate_dirs.extend(cmake_defs['CMAKE_LIBRARY_PATH'].split(';'))
    for dir_ in populate_dirs:
        if not ensure_dir_exists(dir_):
            return False

    # cmake prepares build scripts out-of-source; move into the build output
    # directory and generate scripts from the build directory
    with interim_working_dir(opts.build_output_dir):
        if not CMAKE.execute(cmake_args, env=expand(opts.conf_env)):
            err('failed to prepare cmake project: {}', opts.name)
            return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from releng_tool.packages.pipeline import RelengPackagePipeline
from tests import prepare_testenv
import json
import os
import unittest

class TestPkgPipelineContainedEnvironment(unittest.TestCase):
    def test_pkg_pipeline_containedenv(self):
        pkg_names = [
            'test1',
            'test2',
        ]

        expected_prefix = '/my-custom-prefix'

        with prepare_testenv(template='contained-env') as engine:
            pkgs = engine.pkgman.load(pkg_names)

            pipeline = RelengPackagePipeline(engine, engine.opts, {})
            for pkg in pkgs:
                if not pipeline.process(pkg):
                    break

            # verify package-specific environment variables were set
            self.assertTrue('TEST1_BUILD_OUTPUT_DIR' in os.environ)
            test1_outdir = os.environ['TEST1_BUILD_OUTPUT_DIR']
            results1 = os.path.join(test1_outdir, 'invoke-env.json')
            self.assertTrue(os.path.exists(results1))

            with open(results1, 'r') as f:
                data = json.load(f)

            pkg_keys = [
                'HOST_INCLUDE_DIR',
                'HOST_LIB_DIR',
                'NJOBS',
                'NJOBSCONF',
                'PREFIX',
                'PREFIXED_HOST_DIR',
                'PREFIXED_STAGING_DIR',
                'PREFIXED_TARGET_DIR',
                'STAGING_INCLUDE_DIR',
                'STAGING_LIB_DIR',
                'TARGET_INCLUDE_DIR',
                'TARGET_LIB_DIR',
            ]
            self.assertTrue(all(x in data for x in pkg_keys))
            self.assertEqual(data['NJOBS'], '42')
            self.assertEqual(data['NJOBSCONF'], '42')
            self.assertEqual(data['PREFIX'], expected_prefix)

            opts = engine.opts
            nprefix = os.path.normpath(expected_prefix)
            expected_host_pdir = opts.host_dir + nprefix
            expected_staging_pdir = opts.staging_dir + nprefix
            expected_target_pdir = opts.target_dir + nprefix

            J = os.path.join
            expected_host_include_dir = J(expected_host_pdir, 'include')
            expected_host_lib_dir = J(expected_host_pdir, 'lib')
            expected_staging_include_dir = J(expected_staging_pdir, 'include')
            expected_staging_lib_dir = J(expected_staging_pdir, 'lib')
            expected_target_include_dir = J(expected_target_pdir, 'include')
            expected_target_lib_dir = J(expected_target_pdir, 'lib')

            self.assertEqual(
                data['HOST_INCLUDE_DIR'], expected_host_include_dir)
            self.assertEqual(
                data['HOST_LIB_DIR'], expected_host_lib_dir)
            self.assertEqual(
                data['PREFIXED_HOST_DIR'], expected_host_pdir)
            self.assertEqual(
                data['PREFIXED_STAGING_DIR'], expected_staging_pdir)
            self.assertEqual(
                data['PREFIXED_TARGET_DIR'], expected_target_pdir)
            self.assertEqual(
                data['STAGING_INCLUDE_DIR'], expected_staging_include_dir)
            self.assertEqual(
                data['STAGING_LIB_DIR'], expected_staging_lib_dir)
            self.assertEqual(
                data['TARGET_INCLUDE_DIR'], expected_target_include_dir)
            self.assertEqual(
                data['TARGET_LIB_DIR'], expected_target_lib_dir)

            # verify package-specific environment variables were properly
            # restored to original counts (leak check)
            self.assertTrue('TEST2_BUILD_OUTPUT_DIR' in os.environ)
            test2_outdir = os.environ['TEST2_BUILD_OUTPUT_DIR']
            results2 = os.path.join(test2_outdir, 'invoke-env.json')
            self.assertTrue(os.path.exists(results2))

            with open(results2, 'r') as f:
                data = json.load(f)

            if 'NJOBS' in data:
                self.assertNotEqual(data['NJOBS'], '42')
            if 'NJOBSCONF' in data:
                self.assertNotEqual(data['NJOBSCONF'], '42')
            if 'PREFIX' in data:
                self.assertNotEqual(data['PREFIX'], '/my-custom-prefix')
                self.assertNotEqual(
                    data['HOST_INCLUDE_DIR'], expected_host_include_dir)
                self.assertNotEqual(
                    data['HOST_LIB_DIR'], expected_host_lib_dir)
                self.assertNotEqual(
                    data['PREFIXED_HOST_DIR'], expected_host_pdir)
                self.assertNotEqual(
                    data['PREFIXED_STAGING_DIR'], expected_staging_pdir)
                self.assertNotEqual(
                    data['PREFIXED_TARGET_DIR'], expected_target_pdir)
                self.assertNotEqual(
                    data['STAGING_INCLUDE_DIR'], expected_staging_include_dir)
                self.assertNotEqual(
                    data['STAGING_LIB_DIR'], expected_staging_lib_dir)
                self.assertNotEqual(
                    data['TARGET_INCLUDE_DIR'], expected_target_include_dir)
                self.assertNotEqual(
                    data['TARGET_LIB_DIR'], expected_target_lib_dir)
